from functools import wraps
from types import FunctionType
from typing import NamedTuple, List

import numpy as np

import Orange


HistoryEntry = NamedTuple("HistoryEntry", (("step", FunctionType),
                                           ("centroids", np.ndarray),
                                           ("clusters", np.ndarray)))


def historic(reassign=True):
    def decorator(f):
        @wraps(f)
        def historian(self, *args, **kwargs):
            # store decorated function (not `f`) to enable comparisons with
            # Kmeans.<whatever>
            self._store_history(historian)
            f(self, *args, **kwargs)
            if reassign:
                self.clusters = self._find_clusters()
        return historian
    return decorator if isinstance(reassign, bool) else decorator(reassign)


class Kmeans:
    """
    K-Means algorithm

    Parameters
    ----------
    data (np.ndarray): two-column d ata used for k-means
    centroids (int, list or numpy.array or int): number of coordinates of centroids
    """
    def __init__(self, data, centroids=3):
        self.data = None
        self.centroids = None
        self.history: List[HistoryEntry] = []
        self.clusters = None
        self.reset(data, centroids)

    @property
    def k(self):
        return len(self.centroids)

    @property
    def waits_reassignment(self):
        return self.history and self.history[-1].step == Kmeans.move_centroids

    @property
    def converged(self):
        """
        Clustering converged if the last three steps were assignment, moving,
        and assignment, with membership assignments being the same.
        """
        if len(self.history) < 3:
            return False
        a, b, c = self.history[-3:]
        return a.step == c.step == Kmeans.assign_membership \
               and b.step == Kmeans.move_centroids \
               and np.all(a.clusters == c.clusters)

    def _find_clusters(self):
        dist = Orange.distance.Euclidean(self.data, self.centroids)
        return np.argmin(dist, axis=1)

    def _store_history(self, step):
        self.history.append(
            HistoryEntry(step, np.copy(self.centroids), np.copy(self.clusters)))

    def step_back(self):
        if self.history:
            _, self.centroids, self.clusters = self.history.pop()

    def reset(self, data, centroids=3):
        self.data = data
        self.centroids = np.array(
                centroids if not isinstance(centroids, int)
                else [self._random_position() for _ in range(centroids)])
        self.clusters = self._find_clusters()
        self.history = []

    @historic
    def assign_membership(self):
        pass

    @historic(reassign=False)
    def move_centroids(self):
        for i in range(self.k):
            points = self.data[self.clusters == i]
            self.centroids[i, :] = np.average(points, axis=0) if points.size \
                                   else self._random_position()

    @historic
    def add_centroid(self, x=None, y=None):
        assert (x is None) == (y is None)
        new = self._random_position() if x is None else np.array([x, y])
        self.centroids = np.vstack((self.centroids, new))

    @historic
    def delete_centroid(self, num):
        self.centroids = np.vstack((self.centroids[:num],
                                    self.centroids[num + 1:]))

    @historic
    def move_centroid(self, _index, x, y):
        self.centroids[_index, :] = np.array([x, y])

    def _random_position(self):
        n = len(self.data)
        sample = np.random.choice(n, np.min((5, n)))
        return np.mean(self.data[sample], axis=0)



import numpy as np
import time
import os

import pyqtgraph as pg

from AnyQt.QtCore import QThread, Qt, pyqtSignal as Signal, QTimer, QPointF
from AnyQt.QtGui import QPen, QFont, QPalette, QColor, QIcon
from AnyQt.QtWidgets import QGraphicsTextItem, QGraphicsRectItem, \
    QGraphicsItemGroup, QSizePolicy, QToolButton

try:
    from AnyQt.QtMultimedia import QSound
    enable_sounds = True

    icons_dir = os.path.join(os.path.split(__file__)[0], "icons")
    speaker_on_icon = QIcon(os.path.join(icons_dir, "speaker-on.svg"))
    speaker_off_icon = QIcon(os.path.join(icons_dir, "speaker-off.svg"))
    del icons_dir
except ImportError:
    enable_sounds = False

from Orange.widgets import gui, settings
from Orange.widgets.utils.colorpalettes import DefaultRGBColors
from Orange.widgets.utils.itemmodels import DomainModel
from Orange.widgets.widget import OWWidget, Msg, Input, Output
from Orange.data import DiscreteVariable, ContinuousVariable, Domain, Table

from orangecontrib.educational.widgets.utils.kmeans import Kmeans


class Autoplay(QThread):
    """
    Class used for separated thread when using "Autoplay" for k-means

    Parameters
    ----------
    owkmeans : OWKmeans
        Instance of OWKmeans class
    """

    def __init__(self, owkmeans):
        QThread.__init__(self)
        self.owkmeans = owkmeans
        self.is_running = True
        self.step_count = 0

    def __del__(self):
        self.wait()

    def stop(self):
        self.is_running = False

    def run(self):
        while (self.is_running and
               not self.owkmeans.k_means.converged and
               self.step_count < 100 and
               self.owkmeans.auto_play_enabled):
            try:
                self.owkmeans.step_trigger.emit()
                self.step_count += 1
            except RuntimeError:
                return
            time.sleep(1.5)
        self.owkmeans.stop_auto_play_trigger.emit()


class AnimateNumpy:
    factors = [0.07, 0.26, 0.52, 0.77, 0.95, 1]

    def __init__(self, start, final, callback, done):
        self.start = start
        self.final = final
        self.diff = final - start
        self.callback = callback
        self.done = done
        self.step = 0

    def __call__(self):
        self.step += 1
        if self.step == len(self.factors):
            self.done(self.final)
        else:
            try:
                self.callback(self.start + self.diff * self.factors[self.step])
            except:
                # this is bad, but move on, otherwise you'll be stuck here forever
                pass


class KMeansPlotWidget(pg.PlotWidget):
    centroid_dragged = Signal(int, float, float)
    centroid_done_dragging = Signal(int, float, float)
    centroid_clicked = Signal(int)
    graph_clicked = Signal(float, float)
    mouse_entered = Signal()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.centroids_item = None
        self.moved = False
        self.mouse_down = False
        self.centroid_index = None

    def set_centroids_item(self, centroids_item):
        self.centroids_item = centroids_item

    def mousePressEvent(self, ev):
        if ev.button() != Qt.LeftButton \
                or self.centroids_item is None:
            ev.ignore()
            return

        ev.accept()
        self.mouse_down = True
        pos = self.plotItem.mapToView(QPointF(ev.pos()))
        pts = self.centroids_item.pointsAt(pos)
        if len(pts) != 0:
            self.centroid_index = \
                self.centroids_item.points().tolist().index(pts[0])

    def mouseReleaseEvent(self, ev):
        if ev.button() != Qt.LeftButton:
            ev.ignore()
            return

        ev.accept()
        pos = self.plotItem.mapToView(QPointF(ev.pos()))
        x, y = pos.x(), pos.y()
        if self.centroid_index is not None:
            if self.moved:
                self.centroid_done_dragging.emit(self.centroid_index, x, y)
            else:
                self.centroid_clicked.emit(self.centroid_index)
        else:
            if not self.moved:
                self.graph_clicked.emit(x, y)
        self.centroid_index = None
        self.mouse_down = False
        self.moved = False

    def mouseMoveEvent(self, ev):
        if not self.mouse_down:
            ev.ignore()
            return

        ev.accept()
        if self.centroid_index is not None:
            self.moved = True
            pos = self.plotItem.mapToView(QPointF(ev.pos()))
            self.centroid_dragged.emit(self.centroid_index, pos.x(), pos.y())

    def enterEvent(self, ev):
        super().enterEvent(ev)
        self.mouse_entered.emit()


class OWKmeans(OWWidget):
    name = "Interactive k-Means"
    description = "Widget demonstrates working of k-means algorithm."
    keywords = ["kmeans", "clustering", "interactive"]
    icon = "icons/InteractiveKMeans.svg"
    want_main_area = True
    want_control_area = False
    priority = 300

    class Inputs:
        data = Input("Data", Table)

    class Outputs:
        annotated_data = Output("Annotated Data", Table, default=True)
        centroids = Output("Centroids", Table)

    class Error(OWWidget.Error):
        num_features = Msg("Data must contain at least two numeric variables.")
        no_nonnan_data = Msg("No points with defined values.")

    settingsHandler = settings.DomainContextHandler()
    attr_x = settings.ContextSetting(None)
    attr_y = settings.ContextSetting(None)
    sound_effects = settings.Setting(False)

    graph_name = 'plot'  # pg.GraphicsItem  (pg.PlotItem)
    move_sound = regroup_sound = None

    step_trigger = Signal()
    stop_auto_play_trigger = Signal()

    def __init__(self):
        super().__init__()
        self.data = None
        self.reduced_data = None  # data from selected columns, no nans
        self.selected_rows = None  # rows without nans, which are used in kmeans
        self.in_animation = False
        self.auto_play_enabled = False
        self.auto_play_thread = None
        self.k_means = None
        self.color_map = np.empty(0, dtype=int)
        self.number_of_clusters = 3

        self._create_variables_box()
        self._create_plot()
        self._create_buttons_box()

        if enable_sounds:
            sound_dir = os.path.join(os.path.split(__file__)[0], "sounds")
            self.regroup_sound = QSound(os.path.join(sound_dir, "clack.wav"))
            self.move_sound = QSound(os.path.join(sound_dir, "arrow.wav"))

    def _simplify_widget(self):
        axes = ("bottom", "left")
        self.variables_box.setVisible(len(self.var_model) > 2)
        if [var.name for var in self.var_model] == ["x", "y"]:
            if np.min(self.reduced_data) >= 0 and np.max(self.reduced_data) <= 1:
                for axis in axes:
                    self.plot.hideAxis(axis)
            else:
                for axis in axes:
                    self.plot.getAxis(axis).showLabel(False)
        else:
            for axis, attr in zip(axes, (self.attr_x, self.attr_y)):
                axis = self.plot.getAxis(axis)
                axis.showLabel(True)
                axis.setLabel(attr.name)

    #########################################
    # Variable combo boxes
    def _create_variables_box(self):
        self.variables_box = gui.hBox(self.mainArea, box=True)
        gui.widgetLabel(self.variables_box, "Variables:")
        self.var_model = DomainModel(
            valid_types=(ContinuousVariable, ),
            order=DomainModel.MIXED)
        for attr in ("attr_x", "attr_y"):
            gui.comboBox(
                self.variables_box, self, value=attr, model=self.var_model,
                callback=self.restart, orientation=Qt.Horizontal,
                sizePolicy=(QSizePolicy.MinimumExpanding, QSizePolicy.Fixed)
            )

    #########################################
    # Buttons: creation and updates, and actions
    def _create_buttons_box(self):
        box = gui.hBox(self.plot_box, spacing=0, margin=0)
        self.step_button = \
            gui.button(box, self, "", callback=self.step)
        self.step_back_button = \
            gui.button(box, self, "Step Back", callback=self.step_back)
        if enable_sounds:
            self.speaker_button = QToolButton(self)
            self.speaker_button.setToolButtonStyle(Qt.ToolButtonIconOnly)
            self._update_sound_effect_button()
            self.speaker_button.clicked.connect(self._toggle_sound_effects)
            self.speaker_button.setStyleSheet("border: none")
            box.layout().addWidget(self.speaker_button)
        gui.rubber(box)
        self.restart_button = \
            gui.button(box, self, "Randomize", callback=self.restart)
        self.auto_play_button = \
            gui.button(box, self, "", callback=self.auto_play)
        self._update_buttons()

    if enable_sounds:
        def _toggle_sound_effects(self):
            self.sound_effects = not self.sound_effects
            self._update_sound_effect_button()

        def _update_sound_effect_button(self):
            self.speaker_button.setIcon(
                QIcon([speaker_off_icon, speaker_on_icon][self.sound_effects]))

    def _update_buttons(self):
        animation = self.in_animation
        no_data = self.data is None or self.k_means is None
        running = self.auto_play_enabled
        disabled = animation or no_data or running

        self.variables_box.setDisabled(disabled)

        self.step_button.setDisabled(disabled)
        self.step_button.setText(
            "Reassign Membership"
            if self.k_means and self.k_means.waits_reassignment
            else "Recompute Centroids")

        history = bool(self.k_means and self.k_means.history)
        self.step_back_button.setVisible(history)
        self.step_back_button.setDisabled(
            disabled
            or history
            and len(self.k_means.history[-1].centroids)
            != self.number_of_clusters)

        self.restart_button.setDisabled(disabled)

        # Stop button is never disabled when running (even during animation)
        self.auto_play_button.setDisabled(disabled and not running)
        self.auto_play_button.setText("Stop" if running else "Run Simulation")

    def step(self):
        if self.k_means.waits_reassignment:
            self.k_means.assign_membership()
            self.animate_membership()
        else:
            self.k_means.move_centroids()
            self.animate_centroids()
        self._update_buttons()
        self.send_data()

    def step_back(self):
        self.k_means.step_back()
        self.animate_centroids()
        self.animate_membership()
        self._update_buttons()
        self.send_data()
        self.number_of_clusters = self.k_means.k

    def auto_play(self):
        self.auto_play_enabled = not self.auto_play_enabled
        self._update_buttons()
        if self.auto_play_enabled:
            # This will disable all except the auto_play button
            self.auto_play_thread = Autoplay(self)
            self.step_trigger.connect(self.step)
            self.stop_auto_play_trigger.connect(self.stop_auto_play)
            self.auto_play_thread.start()
        else:
            self.stop_auto_play()

    def stop_auto_play(self):
        self.auto_play_enabled = False
        self._update_buttons()

    #########################################
    # Plot creation
    def _create_plot(self):
        self.points_item = None
        self.centroids_item = None
        self.lines_item = None

        self.plotview = KMeansPlotWidget(background="w", autoRange=False)
        self.plot = self.plotview.getPlotItem()
        axis_pen = QPen(self.palette().color(QPalette.Text))
        tickfont = QFont(self.font())
        tickfont.setPixelSize(max(int(tickfont.pixelSize() * 2 // 3), 11))
        for axis in ("bottom", "left"):
            axis = self.plot.getAxis(axis)
            axis.setPen(axis_pen)
            axis.setTextPen(axis_pen)
            axis.setTickFont(tickfont)
        self.plot.hideButtons()
        self.plot.setMouseEnabled(x=False, y=False)
        self.plotview.centroid_clicked.connect(self.on_centroid_clicked)
        self.plotview.centroid_dragged.connect(self.on_centroid_dragged)
        self.plotview.centroid_done_dragging.connect(self.on_centroid_done_dragging)
        self.plotview.graph_clicked.connect(self.on_centroid_add)

        self.plot_box = gui.vBox(self.mainArea, box=True)
        self.plot_box.layout().addWidget(self.plotview)

        self._create_tooltip()
        self.plotview.mouse_entered.connect(self.show_tooltip)

    def _create_tooltip(self):
        text = QGraphicsTextItem()
        text.setHtml("Drag centroids to move them, click to add and remove them.")
        text.setPos(4, 2)
        r = text.boundingRect()
        text.setTextWidth(r.width())
        rect = QGraphicsRectItem(0, 0, r.width() + 8, r.height() + 4)
        rect.setBrush(QColor(224, 224, 224, 212))
        rect.setPen(QPen(Qt.NoPen))

        tooltip_group = QGraphicsItemGroup()
        tooltip_group.addToGroup(rect)
        tooltip_group.addToGroup(text)
        tooltip_group.setFlag(tooltip_group.ItemIgnoresTransformations)
        self.tooltip = tooltip_group

    def show_tooltip(self):
        if self.tooltip is None or self.data is None or self.k_means is None:
            return

        tooltip = self.tooltip
        self.tooltip = None

        self.plot.scene().addItem(tooltip)
        tooltip.setPos(10, 10)
        timer = QTimer(self)
        timer.singleShot(5000, lambda: self.plot.scene().removeItem(tooltip))

    def _set_plot_items(self):
        self._set_membership_lines()
        self._set_points()
        self._set_centroids()

    def _set_membership_lines(self):
        m = np.zeros(2 * len(self.reduced_data))
        self.lines_item = pg.PlotCurveItem(
            x=m, y=m, pen=pg.mkPen(0.5), connect="pairs", antialias=True)
        self.plotview.addItem(self.lines_item)

    def _set_points(self):
        assert self.k_means is not None
        km = self.k_means
        self.points_item = pg.ScatterPlotItem(
            *self.reduced_data.T,
            symbol="o", size=8, antialias=True, useCache=False)
        self.update_membership()
        self.plotview.addItem(self.points_item)
        self.plotview.autoRange()
        self.plotview.replot()

    def _set_centroids(self):
        assert self.k_means is not None
        k = self.k_means.k
        m = np.zeros(k)
        self.centroids_item = pg.ScatterPlotItem(
            x=m, y=m, pen=self.centr_pens[:k], brush=self.brushes[:k],
            symbol="s", size=13, antialias=True, useCache=False)
        self.plotview.set_centroids_item(self.centroids_item)
        self.plotview.addItem(self.centroids_item)
        self.update_centroid_positions()
        self.update_membership_lines()

    #########################################
    # Plot element updates and animations
    def update_plot(self):
        self.update_centroid_positions()
        self.update_membership()

    def update_membership(self):
        self.update_point_colors()
        self.update_membership_lines()

    def update_point_colors(self):
        assert self.points_item is not None
        km = self.k_means
        self.points_item.setPen(self.pens[km.clusters])
        self.points_item.setBrush(self.brushes[km.clusters])

    def update_centroid_positions(self, cx=None, cy=None):
        assert self.centroids_item is not None
        assert self.lines_item is not None

        km = self.k_means
        k = km.k
        if cx is None:
            cx, cy = km.centroids.copy().T
        self.centroids_item.setData(
            cx, cy,
            pen=self.centr_pens[:k], brush=self.brushes[:k])

    def update_membership_lines(self, cx=None, cy=None):
        x, y = self._membership_lines_data(cx, cy)
        self.lines_item.setData(x, y)

    def _membership_lines_data(self, cx=None, cy=None):
        km = self.k_means
        if cx is None:
            cx, cy = km.centroids.T
        n = len(self.reduced_data)
        x = np.empty(2 * n)
        y = np.empty(2 * n)
        x[::2], y[::2] = self.reduced_data.T
        x[1::2] = cx[km.clusters]
        y[1::2] = cy[km.clusters]
        return x, y

    def animate_centroids(self):
        def update(pos):
            self.update_centroid_positions(*pos.T)
            self.update_membership_lines(*pos.T)

        start = np.array(self.centroids_item.getData()).T
        final = self.k_means.centroids.copy()
        self._animate(start, final, update, update)
        if enable_sounds and self.sound_effects:
            self.move_sound.play()

    def animate_membership(self):
        def update(pos):
            self.lines_item.setData(*pos.T)

        def done(pos):
            update(pos)
            self.update_point_colors()

        start = np.array(self.lines_item.getData()).T
        final = np.array(self._membership_lines_data()).T
        diff = np.any((start != final)[1::2], axis=1)
        pens = self.pens[self.k_means.clusters]
        pens[diff] = [pg.mkPen(pen.color(), width=2) for pen in pens[diff]]
        self.points_item.setPen(pens)
        self._animate(start, final, update, done)
        if enable_sounds and self.sound_effects:
            self.regroup_sound.play()

    def _animate(self, start, final, update, done):
        def my_done(pos):
            timer.stop()
            self.in_animation = False
            self._update_buttons()
            done(pos)

        timer = QTimer(self, interval=50)
        self.in_animation = True
        self._update_buttons()
        timer.start()
        timer.timeout.connect(AnimateNumpy(start, final, update, my_done))

    #########################################
    # Plot: user interaction
    def on_centroid_dragged(self, centroid_index, x, y):
        if self.auto_play_enabled:
            return
        centroids = self.k_means.centroids.copy()
        centroids[centroid_index] = [x, y]
        self.update_centroid_positions(*centroids.T)
        self.update_membership_lines(*centroids.T)

    def on_centroid_done_dragging(self, centroid_index, x, y):
        if self.auto_play_enabled:
            return
        self.k_means.move_centroid(centroid_index, x, y)
        self.animate_membership()
        self.send_data()

    def on_centroid_clicked(self, centroid_index):
        if self.number_of_clusters == 1 or self.auto_play_enabled:
            return
        self.k_means.delete_centroid(centroid_index)
        self.color_map = np.hstack((self.color_map[:centroid_index],
                                    self.color_map[centroid_index + 1:]))
        self._set_colors()
        self.number_of_clusters -= 1
        self.update_centroid_positions()
        self.animate_membership()
        self.send_data()

    def on_centroid_add(self, x, y):
        if not self.data \
                or self.k_means is None \
                or self.number_of_clusters == self.max_clusters() \
                or self.auto_play_enabled:
            return

        self.number_of_clusters += 1
        self.color_map = np.hstack(
            (self.color_map,
             [min(set(range(10))- set(self.color_map))]))
        self._set_colors()
        self.k_means.add_centroid(x, y)
        self.update_centroid_positions()
        self.animate_membership()
        self.send_data()

    def _set_colors(self):
        colors = DefaultRGBColors.qcolors[self.color_map]
        self.pens = np.array([pg.mkPen(col.darker(120)) for col in colors])
        self.centr_pens = np.array([pg.mkPen(col.darker(120), width=2) for col in colors])
        self.brushes = np.array([pg.mkBrush(col) for col in colors])

    #########################################
    # Signals reports ...
    @Inputs.data
    def set_data(self, data):
        self.Error.clear()
        self.plotview.clear()
        self._update_buttons()
        if self.auto_play_thread:
            self.auto_play_thread.stop()

        self.data = data
        self.reduced_data = None  # updated below, at restart
        if not data:
            self.var_model.set_domain(None)
            return
        else:
            self.var_model.set_domain(data.domain)
            if len(self.var_model) < 2:
                self.Error.num_features()
                self.var_model.set_domain(None)
                return

        self.attr_x, self.attr_y = self.var_model[:2]
        self.restart()

    def restart(self):
        """Triggered on data change, attribute change or restart button"""
        self.plotview.clear()
        self.Error.no_nonnan_data.clear()
        self.reduced_data = self._prepare_data()
        if self.reduced_data is None:
            self.Error.no_nonnan_data()
            self.k_means = None
        else:
            if self.number_of_clusters > self.max_clusters():
                self.number_of_clusters = self.max_clusters()
            self.k_means = Kmeans(self.reduced_data, self.number_of_clusters)
            self.color_map = np.arange(self.number_of_clusters)
            self._set_colors()
            self._set_plot_items()
            self._simplify_widget()
        self._update_buttons()
        self.send_data()

    def _prepare_data(self):
        """
        Prepare 2d data for clustering

        Put the two columns into a new table and remove rows with nans.
        Return None if there are no non-nan columns
        """
        attrs = [self.attr_x, self.attr_y]
        x = np.vstack(tuple(self.data.get_column(attr) for attr in attrs)).T
        not_nan = ~np.isnan(x).any(axis=1)
        x = x[not_nan]  # remove rows with nan
        if not x.size:
            return None
        self.selected_rows = np.where(not_nan)
        return x

    def max_clusters(self):
        if self.reduced_data is None:
            return 10
        return min(10, len(self.reduced_data))

    def send_data(self):
        km = self.k_means
        if km is None or km.clusters is None:
            self.Outputs.annotated_data.send(None)
            self.Outputs.centroids.send(None)
        else:
            clust_var = DiscreteVariable(
                "cluster",
                values=["C%d" % (x + 1) for x in range(km.k)],
            )
            clust_var.colors = DefaultRGBColors.palette[self.color_map]
            attributes = self.data.domain.attributes
            classes = self.data.domain.class_vars
            meta_attrs = self.data.domain.metas
            if classes:
                meta_attrs += classes
            classes = [clust_var]
            domain = Domain(attributes, classes, meta_attrs)
            annotated_data = Table.from_table(domain, self.data)
            with annotated_data.unlocked(annotated_data.Y):
                annotated_data.Y[self.selected_rows] = km.clusters

            if self.attr_x is self.attr_y:
                attrs = [self.attr_x.renamed(name) for name in "xy"]
            else:
                attrs = [self.attr_x, self.attr_y]
            centroids = Table.from_numpy(Domain(attrs), km.centroids.copy())
            self.Outputs.annotated_data.send(annotated_data)
            self.Outputs.centroids.send(centroids)

    def send_report(self):
        if self.data is None:
            return
        self.report_plot()


if __name__ == "__main__":
    from orangewidget.utils.widgetpreview import WidgetPreview
    WidgetPreview(OWKmeans).run(Table.from_file('iris'))



import unittest

import numpy as np
from Orange.data import (
    Table,
    DiscreteVariable,
    ContinuousVariable,
    TimeVariable,
)
from Orange.widgets.tests.base import WidgetTest

from orangecontrib.educational.widgets.owcreatetable import OWCreateTable


class TestOWCreateTable(WidgetTest):
    def setUp(self):
        self.widget: OWCreateTable = self.create_widget(OWCreateTable)

    def set_data(self, data):
        self.widget.table_model.set_table(data)
        self.widget.data_changed()

    def test_output(self):
        output = self.get_output(self.widget.Outputs.data)
        # by default widget table have size 3x3 and is empty
        self.assertEqual(3, len(output))
        np.testing.assert_array_equal(np.zeros((3, 3)) * np.nan, output.X)
        self.assertListEqual(
            ["1", "2", "3"], [a.name for a in output.domain.attributes]
        )

        _input = [["1", "2", "3"], ["4", "5", "6"]]
        self.set_data(_input)
        self.wait_until_finished()
        output = self.get_output(self.widget.Outputs.data)
        self.assertEqual(2, len(output))
        np.testing.assert_array_equal(np.array(_input, dtype=float), output.X)

    def test_discrete_columns(self):
        """
        Test different combinations that would result in discrete columns
        """
        _input = [["a", "c", "a"], ["b", "d", "2020-01-01"], ["b", "5", "6"]]
        self.set_data(_input)
        self.wait_until_finished()
        output = self.get_output(self.widget.Outputs.data)

        self.assertEqual(3, len(output.domain.attributes))
        correct_values = [
            {"a", "b"},
            {"d", "c", "5"},
            {"a", "2020-01-01", "6"},
        ]
        for a, cv in zip(output.domain.attributes, correct_values):
            self.assertIsInstance(a, DiscreteVariable)
            self.assertSetEqual(cv, set(a.values))

    def test_continuous_columns(self):
        _input = [["1", "2.0", "3"], ["4", "5.0", "6"]]
        self.set_data(_input)
        self.wait_until_finished()
        output = self.get_output(self.widget.Outputs.data)

        self.assertEqual(3, len(output.domain.attributes))
        for a in output.domain.attributes:
            self.assertIsInstance(a, ContinuousVariable)

    def test_time_columns(self):
        _input = [
            [
                "2020-01-01",
                "2020-10-10T12:08:51",
                "2020-10-10 12:08:51",
                "2020-10-10 12:08",
            ],
            [
                "2020-01-02",
                "2020-10-10T12:09:51",
                "2020-10-10 12:09:51",
                "2020-10-10 12:09",
            ],
            [
                "2020-01-03",
                "2020-10-10T12:10:51",
                "2020-10-10 12:10:51",
                "2020-10-10 12:10",
            ],
        ]
        self.set_data(_input)
        self.wait_until_finished()
        output = self.get_output(self.widget.Outputs.data)

        self.assertEqual(4, len(output.domain.attributes))
        for a in output.domain.attributes:
            self.assertIsInstance(a, TimeVariable)

    def test_none_in_table(self):
        _input = [
            ["2020-01-01", "a", "1", None],
            ["2020-01-02", "b", "2", None],
            [None, None, None, None],
        ]
        self.set_data(_input)
        self.wait_until_finished()
        output = self.get_output(self.widget.Outputs.data)

        correct_types = [
            TimeVariable,
            DiscreteVariable,
            ContinuousVariable,
            ContinuousVariable,
        ]
        for a, typ in zip(output.domain.attributes, correct_types):
            self.assertIsInstance(a, typ)

    def test_num_col_row_change(self):
        _input = [["1", "2", "3"], ["4", "5", "6"], ["7", "8", "9"]]
        self.set_data(_input)
        self.wait_until_finished()

        self.widget.r_spin.setValue(2)
        output = self.get_output(self.widget.Outputs.data)
        self.assertEqual(3, len(output.domain.variables))
        self.assertEqual(2, len(output))
        self.assertEqual(2, self.widget.table_model.rowCount())
        self.assertEqual(3, self.widget.table_model.columnCount())

        self.widget.c_spin.setValue(2)
        output = self.get_output(self.widget.Outputs.data)
        self.assertEqual(2, len(output.domain.variables))
        self.assertEqual(2, len(output))
        self.assertEqual(2, self.widget.table_model.rowCount())
        self.assertEqual(2, self.widget.table_model.columnCount())
        np.testing.assert_array_equal(
            np.array(_input, dtype=float)[:2, :2], output.X
        )

        self.widget.r_spin.setValue(3)
        output = self.get_output(self.widget.Outputs.data)
        self.assertEqual(2, len(output.domain.variables))
        self.assertEqual(3, len(output))
        self.assertEqual(3, self.widget.table_model.rowCount())
        self.assertEqual(2, self.widget.table_model.columnCount())

        self.widget.c_spin.setValue(3)
        output = self.get_output(self.widget.Outputs.data)
        self.assertEqual(3, len(output.domain.variables))
        self.assertEqual(3, len(output))
        self.assertEqual(3, self.widget.table_model.rowCount())
        self.assertEqual(3, self.widget.table_model.columnCount())
        np.testing.assert_array_equal(
            np.hstack(
                (
                    np.vstack(
                        (
                            np.array(_input, dtype=float)[:2, :2],
                            np.zeros((1, 2)) * np.nan,
                        )
                    ),
                    np.zeros((3, 1)) * np.nan,
                )
            ),
            output.X,
        )

    def test_domain(self):
        iris = Table("iris")

        self.send_signal(self.widget.Inputs.data, iris)
        self.wait_until_finished()
        output = self.get_output(self.widget.Outputs.data)

        self.assertEqual(3, self.widget.table_model.rowCount())
        self.assertEqual(5, self.widget.table_model.columnCount())
        self.assertEqual(5, len(output.domain.variables))
        self.assertEqual(3, len(output))

        np.testing.assert_array_equal(np.zeros((3, 4)) * np.nan, output.X)
        np.testing.assert_array_equal(np.zeros((3,)) * np.nan, output.Y)

    def test_context(self):
        iris = Table("iris")

        input1_ = [["1", "2", "3"], ["4", "5", "6"], ["7", "8", "9"]]
        self.set_data(input1_)
        self.wait_until_finished()
        output = self.get_output(self.widget.Outputs.data)

        np.testing.assert_array_equal(np.array(input1_, dtype=float), output.X)

        self.send_signal(self.widget.Inputs.data, iris)
        input2_ = [
            ["1", "2", "3", "4", "Iris-setosa"],
            ["4", "5", "6", "8", "Iris-setosa"],
            ["7", "8", "9", "10", "Iris-setosa"],
        ]
        self.set_data(input2_)
        self.wait_until_finished()
        output = self.get_output(self.widget.Outputs.data)

        np.testing.assert_array_equal(
            np.array(np.array(input2_)[:, :4], dtype=float), output.X
        )

        self.send_signal(self.widget.Inputs.data, None)
        self.wait_until_finished()
        output = self.get_output(self.widget.Outputs.data)
        np.testing.assert_array_equal(np.array(input1_, dtype=float), output.X)

        self.send_signal(self.widget.Inputs.data, iris)
        self.wait_until_finished()
        output = self.get_output(self.widget.Outputs.data)
        np.testing.assert_array_equal(
            np.array(np.array(input2_)[:, :4], dtype=float), output.X
        )


if __name__ == "__main__":
    unittest.main()



import copy
from colorsys import rgb_to_hsv, hsv_to_rgb
from itertools import chain

import numpy as np
from scipy.interpolate import splprep, splev
from scipy.ndimage.filters import gaussian_filter

from AnyQt.QtCore import Qt, QRectF, QObject, QPointF, QEvent
from AnyQt.QtGui import QPalette, QPen, QFont, QCursor, QColor, QBrush
from AnyQt.QtWidgets import QGraphicsSceneMouseEvent, QGraphicsTextItem

import pyqtgraph as pg

from Orange.data import \
    Table, Domain, ContinuousVariable, StringVariable, DiscreteVariable
from Orange.base import Learner
from Orange.classification import \
    LogisticRegressionLearner, RandomForestLearner, TreeLearner
from Orange.preprocess.transformation import Indicator

from Orange.widgets import gui
from Orange.widgets.settings import \
    DomainContextHandler, Setting, SettingProvider, ContextSetting
from Orange.widgets.utils.colorpalettes import DiscretePalette
from Orange.widgets.utils.itemmodels import DomainModel
from Orange.widgets.utils.owlearnerwidget import OWBaseLearner
from Orange.widgets.visualize.owscatterplotgraph import OWScatterPlotBase
from Orange.widgets.widget import Msg, Input, Output

from orangecontrib.educational.widgets.utils.gradient_grid import \
    interpolate_grid
from orangecontrib.educational.widgets.utils.polynomialtransform \
    import PolynomialTransform
from orangecontrib.educational.widgets.utils.contour import Contour

GRID_SIZE = 60


class HoverEventDelegate(QObject):
    def __init__(self, delegate, parent=None):
        super().__init__(parent)
        self.delegate = delegate

    def eventFilter(self, obj, event):
        return isinstance(event, QGraphicsSceneMouseEvent) \
               and self.delegate(event)


class PolynomialPlot(OWScatterPlotBase):
    def __init__(self, scatter_widget, parent=None):
        super().__init__(scatter_widget, parent)
        self.alpha_value = 255

    def _get_discrete_colors(self, c_data, subset):
        pen, _ = super()._get_discrete_colors(c_data, subset)

        probs = self.master.get_probabilities()
        if probs is None:
            return pen, QBrush(self.palette.qcolors_w_nan[-1])

        probs = probs[:, 1]
        palette = self.master.get_model_palette().palette
        colors = np.hstack((palette[(probs > 0.5).astype(int)],
                            128 + 255 * np.abs(probs - 0.5)[:, None].astype(int)))
        brush = [QBrush(QColor(*[int(x) for x in col])) for col in colors]
        return pen, brush


class OWPolynomialClassification(OWBaseLearner):
    name = "Polynomial Classification"
    description = "Widget that demonstrates classification " \
                  "with polynomial expansion of variables."
    keywords = ["polynomial classification", "classification", "class",
                "classification visualization", "polynomial features"]
    icon = "icons/polynomialclassification.svg"
    want_main_area = True
    resizing_enabled = True
    priority = 600

    class Inputs(OWBaseLearner.Inputs):
        learner = Input("Learner", Learner)

    class Outputs(OWBaseLearner.Outputs):
        coefficients = Output("Coefficients", Table, default=True)
        data = Output("Data", Table)

    LEARNER = LogisticRegressionLearner

    settingsHandler = DomainContextHandler(
        match_values=DomainContextHandler.MATCH_VALUES_CLASS)
    graph = SettingProvider(PolynomialPlot)

    learner_name = Setting("Polynomial Classification")
    attr_x = ContextSetting(None)
    attr_y = ContextSetting(None)
    target_class = Setting("")
    degree = Setting(1)
    contours_enabled = Setting(True)

    graph_name = 'graph.plot_widget'  # QGraphicsView (pg.PlotWidget)

    class Error(OWBaseLearner.Error):
        num_features = Msg("Data must contain at least two numeric variables.")
        no_class = Msg("Data must have a single target attribute.")
        no_nonnan_data = Msg("No points with defined values.")
        same_variable = Msg("Select two different variables.")

    def __init__(self, *args, **kwargs):
        # Some attributes must be created before super().__init__
        self._add_graph()
        self.var_model = DomainModel(valid_types=(ContinuousVariable, ))

        super().__init__(*args, **kwargs)
        self.input_learner = None
        self.data = None

        self.learner = None
        self.selected_data = None
        self.orig_class = None
        self.probabilities_grid = None
        self.xv = None
        self.yv = None
        self.contours = []
        self.init_learner()

    def add_main_layout(self):
        box = gui.widgetBox(self.controlArea, "Variables")
        gui.comboBox(
            box, self, "attr_x", model=self.var_model,
            callback=self._on_attr_changed)
        gui.comboBox(
            box, self, "attr_y", model=self.var_model,
            callback=self._on_attr_changed)
        gui.spin(
            box, self, value='degree', label='Polynomial expansion:',
            minv=1, maxv=5, step=1, alignment=Qt.AlignRight, controlWidth=70,
            callback=self._on_degree_changed)
        gui.comboBox(
            box, self, 'target_class', label='Target:',
            orientation=Qt.Horizontal, sendSelectedValue=True,
            callback=self._on_target_changed)

        box = gui.widgetBox(self.controlArea, box=True)
        gui.checkBox(
            box, self.graph, 'show_legend',"Show legend",
            callback=self.graph.update_legend_visibility)
        gui.checkBox(
            box, self, 'contours_enabled', label="Show contours",
            callback=self.plot_contour)

        gui.rubber(self.controlArea)

    def add_bottom_buttons(self):
        pass

    def _on_degree_changed(self):
        self.init_learner()
        self.apply()

    def _on_attr_changed(self):
        self.select_data()
        self.apply()

    def _on_target_changed(self):
        self.select_data()
        self.apply()

    ##############################
    # Input signal-related stuff
    @Inputs.learner
    def set_learner(self, learner):
        self.input_learner = learner
        self.init_learner()

    def set_preprocessor(self, preprocessor):
        self.preprocessors = [preprocessor] if preprocessor else []
        self.init_learner()

    @Inputs.data
    def set_data(self, data):
        combo = self.controls.target_class

        self.closeContext()
        self.Error.clear()
        combo.clear()
        self.var_model.set_domain(None)
        self.data = self.selected_data = self.orig_class = None
        self.xv = None
        self.yv = None
        self.probabilities_grid = None

        if not data:
            return
        domain = data.domain
        if domain.class_var is None or domain.class_var.is_continuous:
            self.Error.no_class()
            return
        if sum(var.is_continuous
                 for var in chain(domain.variables, domain.metas)) < 2:
            self.Error.num_features()
            return

        self.data = data
        non_empty = np.bincount(data.Y[np.isfinite(data.Y)].astype(int),
                                minlength=len(domain.class_var.values)) > 0
        values = np.array(domain.class_var.values)[non_empty]
        combo.addItems(values.tolist())
        self.var_model.set_domain(self.data.domain)
        self.attr_x, self.attr_y = self.var_model[:2]
        self.target_class = combo.itemText(0)

        hide_attrs = len(self.var_model) == 2
        self.controls.attr_x.setHidden(hide_attrs)
        self.controls.attr_y.setHidden(hide_attrs)

        self.openContext(self.data)
        self.select_data()

    def init_learner(self):
        self.learner = copy.copy(self.input_learner
                                 or self.LEARNER(penalty='l2', C=1e10))
        self.learner.preprocessors = (
            [PolynomialTransform(self.degree)] +
            list(self.preprocessors or []) +
            list(self.learner.preprocessors or []))
        self.send_learner()

    def handleNewSignals(self):
        self.apply()

    def select_data(self):
        """Put the two selected columns in a new Orange.data.Table"""
        self.Error.no_nonnan_data.clear()
        self.Error.same_variable.clear()

        attr_x, attr_y = self.attr_x, self.attr_y
        if self.attr_x is self.attr_y:
            self.selected_data = None
            self.orig_class = None
            self.Error.same_variable()
            return

        names = [var.name for var in (attr_x, attr_y)]
        if names == ["x", "y"]:
            names = [""] * 2
        for place, name in zip(("bottom", "left"), names):
            self.graph.plot_widget.getAxis(place).setLabel(name)
        old_class = self.data.domain.class_var
        values = old_class.values
        target_idx = values.index(self.target_class)

        binary = len(values) == 2
        new_class = DiscreteVariable(
            old_class.name + "'",
            values=(values[1 - target_idx] if binary else 'Others',
                    self.target_class),
            compute_value=Indicator(old_class, target_idx))
        new_class.palette = DiscretePalette(
            "indicator", "indicator",
            [list(old_class.palette.palette[1 - target_idx])
             if binary else [64, 64, 64],
             list(old_class.palette.palette[target_idx])])

        domain = Domain([attr_x, attr_y], new_class, [old_class])

        self.selected_data = self.data.transform(domain)
        valid_data = \
            np.flatnonzero(
                np.all(
                    np.isfinite(self.selected_data.X),
                    axis=1)
            )
        if not np.any(np.isfinite(self.data.Y[valid_data])):
            self.Error.no_nonnan_data()
            self.selected_data = None
            self.orig_class = None
        else:
            self.selected_data = self.selected_data[valid_data]
            self.orig_class = self.data.Y[valid_data]

    def apply(self):
        self.update_model()
        self.send_model()
        self.send_coefficients()
        self.send_data()

        self.graph.reset_graph()
        self.graph.plot_widget.addItem(self.contour_label)  # Re-add the label
        self.plot_gradient()
        self.plot_contour()

    def update_model(self):
        self.Error.fitting_failed.clear()
        self.model = None
        self.probabilities_grid = None
        if self.selected_data is not None and self.learner is not None:
            try:
                defined = self.selected_data[np.isfinite(self.selected_data.Y)]
                self.model = self.learner(defined)
                self.model.name = self.learner_name
            except Exception as e:
                self.Error.fitting_failed(str(e))

    ##############################
    # Graph and its contents
    def _add_graph(self):
        self.graph = PolynomialPlot(self)
        self.graph.plot_widget.setCursor(QCursor(Qt.CrossCursor))
        self.mainArea.layout().addWidget(self.graph.plot_widget)
        self.graph.point_width = 1

        axis_color = self.palette().color(QPalette.Text)
        axis_pen = QPen(axis_color)

        tickfont = QFont(self.font())
        tickfont.setPixelSize(max(int(tickfont.pixelSize() * 2 // 3), 11))

        for pos in ("bottom", "left"):
            axis = self.graph.plot_widget.getAxis(pos)
            axis.setPen(axis_pen)
            axis.setTickFont(tickfont)
            axis.show()

        self._hover_delegate = HoverEventDelegate(self.help_event)
        self.graph.plot_widget.scene().installEventFilter(self._hover_delegate)

        self.contour_label = label = QGraphicsTextItem()
        label.setFlag(QGraphicsTextItem.ItemIgnoresTransformations)
        self.graph.plot_widget.addItem(label)
        label.hide()
        # I'm not proud of this and will brew a coffee to the person who
        # improves it (and comes to claim it)
        self.graph.plot_widget.scene().leaveEvent = lambda *_: label.hide()

    @staticmethod
    def _contour_pen(value, hovered):
        return pg.mkPen(0.2,
            width=1 + 2 * hovered + (value == 0.5),
            style=Qt.SolidLine)
            # Alternative:
            # Qt.SolidLine if hovered or value == 0.5 else Qt.DashDotLine

    def help_event(self, event):
        if self.probabilities_grid is None:
            return False

        pos = event.scenePos()
        pos = self.graph.plot_widget.mapToView(pos)

        # The mouse hover width is a bit larger for easier hovering, but
        # consequently more than one line can be hovered. Pick the middle one.
        hovereds = [item for item in self.contours
                    if item.mouseShape().contains(pos)]
        hovered = hovereds[len(hovereds) // 2] if hovereds else None
        # Set the pen for all lines, hovered and not hovered (any longer)
        for item in self.contours:
            item.setPen(self._contour_pen(item.value, item is hovered))

        # Set the probability for the textitem at mouse position.
        # If there is a hovered line - this acts as a line label.
        # Otherwise, take the probability from the grid
        label = self.contour_label
        if hovered:
            prob = hovered.value
        else:
            min_x, step_x, min_y, step_y = self.probabilities_grid_dimensions
            prob = self.probabilities_grid.T[
                int(np.clip(round((pos.x() - min_x) * step_x), 0, GRID_SIZE - 1)),
                int(np.clip(round((pos.y() - min_y) * step_y), 0, GRID_SIZE - 1))]
        prob_lab = f"{round(prob, 3):.3g}"
        if "." not in prob_lab:
            prob_lab += ".0"  # Showing just 0 or 1 looks ugly
        label.setHtml(prob_lab)
        font = label.font()
        font.setBold(hovered is not None)
        label.setFont(font)

        # Position the label above and left from mouse; looks nices
        rect = label.boundingRect()
        spos = event.scenePos()
        x, y = spos.x() - rect.width(), spos.y() - rect.height()
        label.setPos(self.graph.plot_widget.mapToView(QPointF(x, y)))

        label.show()
        return True

    def changeEvent(self, ev):
        # This hides the label if the user alt-tabs out of the window
        if ev.type() == QEvent.ActivationChange and not self.isActiveWindow():
            self.contour_label.hide()
        super().changeEvent(ev)

    def plot_gradient(self):
        if not self.model:
            self.probabilities_grid = None
            return

        (min_x, max_x), (min_y, max_y) = self.graph.view_box.viewRange()
        x = np.linspace(min_x, max_x, GRID_SIZE)
        y = np.linspace(min_y, max_y, GRID_SIZE)
        self.xv, self.yv = np.meshgrid(x, y)

        attr = np.hstack((self.xv.reshape((-1, 1)), self.yv.reshape((-1, 1))))
        nil = np.full((len(attr), 1), np.nan)
        attr_data = Table.from_numpy(self.selected_data.domain, attr, nil, nil)

        self.probabilities_grid = self.model(attr_data, 1)[:, 1] \
            .reshape(self.xv.shape)
        self.probabilities_grid_dimensions = (
            min_x, GRID_SIZE / (max_x - min_x),
            min_y, GRID_SIZE / (max_y - min_y))

        if not self._treelike:
            self.probabilities_grid = self.blur_grid(self.probabilities_grid)

        bitmap = interpolate_grid(self.probabilities_grid, 256)
        bitmap *= 255
        bitmap = bitmap.astype(np.uint8)

        class_var = self.selected_data.domain.class_var
        h1, s1, v1 = rgb_to_hsv(*class_var.colors[1] / 255)
        palette = np.vstack((
            np.linspace([h1, 0, 0.8], [h1, 0, 1], 128),
            np.linspace([h1, 0, 1], [h1, s1 * 0.5, 0.7 + 0.3 * v1], 128)
        ))
        palette = 255 * np.array([hsv_to_rgb(*col) for col in palette])
        palette = palette.astype(int)

        density_img = pg.ImageItem(bitmap.T, lut=palette)

        density_img.setRect(QRectF(min_x, min_y,
                                   max_x - min_x, max_y - min_y))
        density_img.setZValue(-1)
        self.graph.plot_widget.addItem(density_img, ignoreBounds=True)

    def remove_contours(self):
        while self.contours:
            self.graph.plot_widget.removeItem(self.contours.pop())

    @property
    def _treelike(self):
        return isinstance(self.learner, (RandomForestLearner, TreeLearner))

    def plot_contour(self):
        self.remove_contours()
        if self.probabilities_grid is None or not self.contours_enabled:
            return

        contour = Contour(self.xv, self.yv, self.probabilities_grid)
        contour_lines = contour.contours(np.arange(0.1, 1, 0.1))
        for key, value in contour_lines.items():
            for line in value:
                if len(line) > self.degree and not self._treelike:
                    tck, u = splprep(
                        [list(x) for x in zip(*reversed(line))],
                        s=0.001, k=self.degree,
                        per=(len(line) if np.allclose(line[0], line[-1])
                             else 0))
                    new_int = np.arange(0, 1.01, 0.01)
                    interpol_line = np.array(splev(new_int, tck)).T.tolist()
                else:
                    interpol_line = line
                contour = pg.PlotCurveItem(
                    *np.array(list(interpol_line)).T,
                    pen=self._contour_pen(key, False))
                # The hover region can be narrowed by calling setClickable
                # (with False, to keep it unclickable)
                contour.value = key
                self.graph.plot_widget.addItem(contour)
                self.contours.append(contour)

    @staticmethod
    def blur_grid(grid):
        filtered = gaussian_filter(grid, sigma=1)
        filtered[(grid > 0.45) & (grid < 0.55)] = grid[(grid > 0.45) &
                                                       (grid < 0.55)]
        return filtered

    # The following methods are required by OWScatterPlotBase
    def get_coordinates_data(self):
        if not self.selected_data:
            return None, None
        return self.selected_data.X.T

    def get_color_data(self):
        return self.orig_class

    def get_palette(self):
        return self.data.domain.class_var.palette

    def get_color_labels(self):
        return self.selected_data and self.data.domain.class_var.values

    def is_continuous_color(self):
        return False

    def get_model_palette(self):
        return self.model and self.selected_data.domain.class_var.palette

    def get_probabilities(self):
        return self.model and self.model(self.selected_data, 1)

    get_size_data = get_shape_data = get_shape_labels = \
        get_subset_mask = get_label_data = get_tooltip = selection_changed = \
        lambda *_: None

    ##############################
    # Output signal-related stuff
    def send_learner(self):
        if self.learner is not None:
            self.learner.name = self.learner_name
        self.Outputs.learner.send(self.learner)

    def send_model(self):
        self.Outputs.model.send(self.model)

    def send_coefficients(self):
        if (self.model is None
                or not isinstance(self.learner, LogisticRegressionLearner)
                or not hasattr(self.model, 'skl_model')):
            self.Outputs.coefficients.send(None)
            return

        model = self.model.skl_model
        domain = Domain(
            [ContinuousVariable("coef")], metas=[StringVariable("name")])
        coefficients = model.intercept_.tolist() + model.coef_[0].tolist()
        names = ["Intercept"] + [x.name for x in self.model.domain.attributes]
        coefficients_table = Table.from_list(
            domain, list(zip(coefficients, names)))
        self.Outputs.coefficients.send(coefficients_table)

    def send_data(self):
        if self.selected_data is None:
            self.Outputs.data.send(None)
        else:
            expanded = PolynomialTransform(self.degree)(self.selected_data)
            self.Outputs.data.send(expanded)

    def send_report(self):
        if self.selected_data is None:
            return
        name = "" if self.degree == 1 \
            else f"Model with polynomial expansion {self.degree}"
        self.report_plot(name=name)


if __name__ == "__main__":
    from orangewidget.utils.widgetpreview import WidgetPreview
    WidgetPreview(OWPolynomialClassification).run(Table.from_file('iris'))



import numpy as np
from scipy.interpolate import RegularGridInterpolator


def interpolate_grid(values, resolution):
    """
    Resize grid of values from its size to the shape resolution x resolution.
    Use linear interpolation for points in between current points.

    Parameters
    ----------
    values
        Two-dimensional grid of values
    resolution
        Resolution of resulting grid

    Returns
    -------
    The output grid with shape resolution x resolution
    """
    x = np.linspace(0, 1, values.shape[1])
    y = np.linspace(0, 1, values.shape[0])
    interpolator = RegularGridInterpolator((x, y), values)

    points = np.linspace(0, 1, resolution)
    xv, yv = np.meshgrid(points, points)
    return interpolator((yv, xv))



import unittest
from unittest.mock import Mock

import numpy as np

from Orange.widgets.tests.base import WidgetTest
from Orange.data import Domain, ContinuousVariable, Table

from orangecontrib.educational.widgets import owkmeans
from orangecontrib.educational.widgets.owkmeans import OWKmeans

# No animations
owkmeans.AnimateNumpy.__call__ = lambda self: self.done(self.final)


class TestOWKmeans(WidgetTest):
    def setUp(self):
        self.widget = self.create_widget(OWKmeans)  # type: OWKmeans
        self.data = Table("iris")
        domain = Domain(attributes=self.data.domain.attributes[:2],
                        class_vars=self.data.domain.class_vars)
        self.data2 = self.data.transform(domain)
        domain = Domain([ContinuousVariable(n) for n in "xy"])
        self.dataxy = Table.from_numpy(domain, np.arange(20).reshape(10, 2))

    def test_simplify_widget(self):
        def on_iris(nattrs):
            self.assertIs(w.attr_x, self.data.domain[0])
            self.assertIs(w.attr_y, self.data.domain[1])
            self.assertIs(w.variables_box.isHidden(), nattrs <= 2)
            axes = ("bottom", "left")
            for axis, attr in zip(axes, (w.attr_x, w.attr_y)):
                axis = w.plot.getAxis(axis)
                self.assertTrue(axis.label.isVisible())
                self.assertTrue(axis.label.toPlainText(), attr.name)

        w = self.widget
        self.send_signal(w.Inputs.data, self.data)
        on_iris(4)

        self.send_signal(w.Inputs.data, self.data2)
        on_iris(2)

        self.send_signal(w.Inputs.data, self.data)
        on_iris(4)

        self.send_signal(w.Inputs.data, self.dataxy)
        self.assertIs(w.attr_x, self.dataxy.domain[0])
        self.assertIs(w.attr_y, self.dataxy.domain[1])
        self.assertTrue(w.variables_box.isHidden())
        axes = ("bottom", "left")
        for axis, attr in zip(axes, (w.attr_x, w.attr_y)):
            axis = w.plot.getAxis(axis)
            self.assertFalse(axis.label.isVisible())

        self.send_signal(w.Inputs.data, self.data)
        on_iris(4)

    def test_step(self):
        w = self.widget
        self.send_signal(w.Inputs.data, self.data)
        w.send_data = Mock()
        w._update_buttons = Mock()

        kmeans = w.k_means
        klass = type(w.k_means)
        w.step()
        w.send_data.assert_called()
        w._update_buttons.assert_called()
        w.send_data.reset_mock()
        w._update_buttons.reset_mock()
        self.assertIs(kmeans.history[-1].step, klass.move_centroids)

        w.step()
        w.send_data.assert_called()
        w._update_buttons.assert_called()
        w.send_data.reset_mock()
        w._update_buttons.reset_mock()
        self.assertIs(kmeans.history[-1].step, klass.assign_membership)

    def test_step_back(self):
        w = self.widget
        self.send_signal(w.Inputs.data, self.data)
        kmeans = w.k_means

        centroids0 = kmeans.centroids.copy()
        clusters0 = kmeans.clusters.copy()

        w.step()
        centroids1 = kmeans.centroids.copy()
        clusters1 = kmeans.clusters.copy()

        w.step()

        assert not np.all(centroids0 == centroids1)
        w.send_data = Mock()
        w._update_buttons = Mock()

        w.step_back()
        w.send_data.assert_called()
        w._update_buttons.assert_called()
        w.send_data.reset_mock()
        w._update_buttons.reset_mock()
        np.testing.assert_equal(kmeans.centroids, centroids1)
        np.testing.assert_equal(kmeans.clusters, clusters1)

        w.step_back()
        w.send_data.assert_called()
        w._update_buttons.assert_called()
        w.send_data.reset_mock()
        w._update_buttons.reset_mock()
        np.testing.assert_equal(kmeans.centroids, centroids0)
        np.testing.assert_equal(kmeans.clusters, clusters0)

    def test_autoplay_buttons(self):
        """
        Check if everything goes correct for autorun
        """
        self.widget.set_data(self.data)
        self.widget.number_of_clusters = 4

        # start autoplay
        self.widget.auto_play_button.click()

        self.assertEqual(self.widget.variables_box.isEnabled(), False)
        self.assertEqual(self.widget.restart_button.isEnabled(), False)
        self.assertEqual(self.widget.step_button.isEnabled(), False)
        self.assertEqual(self.widget.step_back_button.isEnabled(), False)
        self.assertEqual(self.widget.auto_play_button.isEnabled(), True)
        self.assertEqual(self.widget.auto_play_button.text(), "Stop")

        # stop autoplay
        self.widget.auto_play_button.click()

        self.assertEqual(self.widget.variables_box.isEnabled(), True)
        self.assertEqual(self.widget.restart_button.isEnabled(), True)
        self.assertEqual(self.widget.step_button.isEnabled(), True)
        self.assertEqual(self.widget.auto_play_button.isEnabled(), True)
        self.assertNotEqual(self.widget.auto_play_button.text(), "Stop")

    def test_centroids_change(self):
        self.widget.number_of_clusters = 4
        self.widget.set_data(self.data)
        self.assertEqual(self.widget.k_means.k, 4)
        self.assertEqual(self.widget.number_of_clusters, 4)
        np.testing.assert_equal(self.widget.color_map, [0, 1, 2, 3])

        self.widget.on_centroid_clicked(1)
        self.assertEqual(self.widget.number_of_clusters, 3)
        np.testing.assert_equal(self.widget.color_map, [0, 2, 3])

        self.widget.on_centroid_add(2, 3)
        self.assertEqual(self.widget.number_of_clusters, 4)
        np.testing.assert_equal(self.widget.color_map, [0, 2, 3, 1])

    def test_no_data(self):
        self.widget.set_data(None)
        self.assertIsNone(self.widget.data)
        self.assertIsNone(self.widget.reduced_data)
        self.assertIsNone(self.widget.k_means)
        self.assertEqual(self.widget.var_model.rowCount(), 0)

        # too few continuous attributes: don't create reduced data and k-means
        domain = Domain(self.data.domain.attributes[:1],
                        self.data.domain.class_var)
        data1 = Table.from_table(domain, self.data)
        self.widget.set_data(data1)
        self.assertIsNone(self.widget.reduced_data)
        self.assertIsNone(self.widget.k_means)
        self.assertTrue(self.widget.Error.num_features.is_shown())

    def test_data_no_non_nan_data(self):
        with self.dataxy.unlocked():
            self.dataxy[:5, 0] = np.nan
            self.dataxy[5:, 1] = np.nan
        self.send_signal(self.widget.Inputs.data, self.dataxy)
        self.assertIsNone(self.widget.reduced_data)
        self.assertIsNone(self.widget.k_means)
        self.assertTrue(self.widget.Error.no_nonnan_data.is_shown())

    def test_data_some_nan_data(self):
        with self.dataxy.unlocked():
            self.dataxy[:5, 0] = np.nan
        self.send_signal(self.widget.Inputs.data, self.dataxy)
        np.testing.assert_equal(self.widget.reduced_data, self.dataxy.X[5:])

    def test_restart(self):
        self.widget.set_data(self.data)
        prev_centroids = self.widget.k_means.centroids
        self.widget.restart_button.click()
        self.assertFalse(np.all(prev_centroids - self.widget.k_means.centroids == 0))

    def test_send_report(self):
        w = self.widget
        w.set_data(self.data)
        w.report_button.click()


if __name__ == "__main__":
    unittest.main()



import math

from Orange.data.util import SharedComputeValue
from Orange.evaluation import RMSE, MAE, Results
from AnyQt.QtCore import Qt, QRectF, QPointF
from AnyQt.QtGui import QColor, QPalette, QPen, QFont

import sklearn.preprocessing as skl_preprocessing
import pyqtgraph as pg
import numpy as np

from orangewidget.report import report
from orangewidget.utils.widgetpreview import WidgetPreview

from Orange.data import Table, Domain
from Orange.data.variable import ContinuousVariable, StringVariable
from Orange.regression.linear import RidgeRegressionLearner, LinearRegressionLearner
from Orange.base import Learner
from Orange.regression.mean import MeanModel
from Orange.statistics.distribution import Continuous
from Orange.widgets import settings, gui
from Orange.widgets.widget import Msg, Input, Output
from Orange.widgets.settings import DomainContextHandler
from Orange.widgets.utils.itemmodels import DomainModel
from Orange.widgets.utils.owlearnerwidget import OWBaseLearner
from Orange.widgets.utils.sql import check_sql_input


class PolynomialFeatureSharedCV(SharedComputeValue):
    def __init__(self, compute_shared, idx):
        super().__init__(compute_shared)
        self.idx = idx

    def compute(self, _, shared_data):
        return shared_data[:, self.idx]

    def __eq__(self, other):
        # Remove the first test after we require Orange 3.33
        return type(self) is type(other) \
               and super().__eq__(other) and self.idx == other.idx

    def __hash__(self):
        return hash((super().__hash__(), self.idx))


class PolynomialFeatures:
    def __init__(self, degree=2, include_bias=True):
        self.degree = degree
        self.include_bias = include_bias

    def __call__(self, data):
        if self.degree == 0:
            # Zero degree without intercept shouldn't have a column of 1's,
            # otherwise we do have intercapt. But some data is needed by most
            # learners, so we provide a column of zeros
            variables = [
                ContinuousVariable(
                    "x0",
                    compute_value=TempMeanModel(int(self.include_bias)))]
        else:
            pf = skl_preprocessing.PolynomialFeatures(
                self.degree, include_bias=self.include_bias
            )
            pf.fit(data.X)
            cv = lambda table: pf.transform(table.transform(data.domain).X)
            features = pf.get_feature_names_out() if pf.n_output_features_ else []
            variables = [
                ContinuousVariable(f, compute_value=PolynomialFeatureSharedCV(cv, i))
                for i, f in enumerate(features)
            ]
        domain = Domain(
            variables,
            class_vars=data.domain.class_vars,
            metas=data.domain.metas,
        )
        return data.transform(domain)


class TempMeanModel(MeanModel):
    """
    Using MeanModel Model's __call__ is called that transform table to the
    original domain space which produces empty X - and so the error is raised
    Here we bypass model's __call__
    """
    InheritEq = True

    def __init__(self, const):
        distr = Continuous(np.array([[const], [1.0]]))
        super().__init__(distr)

    def __call__(self, data, *args, **kwargs):
        return self.predict(data)


class RegressTo0(Learner):
    @staticmethod
    def __call__(data, *args, **kwargs):
        model = TempMeanModel(0)
        return model


class PolynomialLearnerWrapper(Learner):
    def __init__(self, x_var, y_var, degree, learner, preprocessors, fit_intercept):
        super().__init__()
        self.x_var = x_var
        self.y_var = y_var
        self.degree = degree
        self.learner = learner
        self.preprocessors = preprocessors
        self.fit_intercept = fit_intercept

    def __call__(self, data: Table, progress_callback=None):
        data = data.transform(Domain([self.x_var], self.y_var))
        *_, model = self.data_and_model(data)
        return model

    def data_and_model(self, data: Table):
        """
        Trains the model, and also returns temporary tables

        The function is used in the widget instead of __call__ to avoid
        recomputing preprocessed and expanded data.
        """
        valid_mask = np.isfinite(data.get_column(self.x_var)) \
                     & np.isfinite(data.get_column(self.y_var))
        data_table = Table.from_table(
            Domain([self.x_var], self.y_var), data[valid_mask]
        )

        # all lines have nan
        if np.all(np.isnan(data_table.X.flatten()) | np.isnan(data_table.Y)):
            return None, None, None, None, None

        # apply preprocessors on the input first
        preprocessed_table = (
            self.preprocessors(data_table) if self.preprocessors else data_table
        )

        # use polynomial preprocessor after applying preprocessors from input
        poly_preprocessor = PolynomialFeatures(
            degree=self.degree, include_bias=self.fit_intercept
        )

        expanded_data = poly_preprocessor(preprocessed_table)
        predictor = self.learner(expanded_data)
        return (data_table, preprocessed_table, poly_preprocessor,
                expanded_data, predictor)


class OWPolynomialRegression(OWBaseLearner):
    name = "Polynomial Regression"
    description = "Univariate regression with polynomial expansion."
    keywords = ["polynomial regression", "regression",
                "regression visualization", "polynomial features"]
    icon = "icons/UnivariateRegression.svg"
    priority = 500

    class Inputs(OWBaseLearner.Inputs):
        learner = Input("Learner", Learner)

    class Outputs(OWBaseLearner.Outputs):
        coefficients = Output("Coefficients", Table, explicit=True)
        data = Output("Data", Table, explicit=True)


    replaces = [
        "Orange.widgets.regression.owunivariateregression."
        "OWUnivariateRegression",
        "orangecontrib.prototypes.widgets.owpolynomialregression.",
        "orangecontrib.educational.widgets.owunivariateregression."
    ]

    LEARNER = LinearRegressionLearner

    learner_name = settings.Setting("Polynomial Regression")

    polynomialexpansion = settings.Setting(1)

    settingsHandler = DomainContextHandler()
    x_var = settings.ContextSetting(None)
    y_var = settings.ContextSetting(None)
    error_bars_enabled = settings.Setting(False)
    fit_intercept = settings.Setting(True)

    default_learner_name = "Linear Regression"
    error_plot_items = []

    rmse = ""
    mae = ""
    regressor_name = ""

    want_main_area = True
    graph_name = 'plot'  # pg.GraphicsItem  (pg.PlotItem)

    class Warning(OWBaseLearner.Warning):
        large_diffs = Msg(
            "Polynomial feature values are very large. "
            "This may cause numerical instabilities."
        )

    class Error(OWBaseLearner.Error):
        all_none = Msg("All rows have undefined data.")
        no_cont_variables =\
            Msg("Regression requires at least two numeric variables.")
        same_dep_indepvar =\
            Msg("Dependent and independent variables must be differnt.")

    def __init__(self):
        super().__init__()
        self.data = None
        self.learner = None

        self.scatterplot_item = None
        self.plot_item = None

    def add_main_layout(self):
        self.rmse = ""
        self.mae = ""
        self.regressor_name = self.default_learner_name

        self.var_model = DomainModel(
            valid_types=(ContinuousVariable, ),
            order=DomainModel.MIXED)

        box = gui.vBox(self.controlArea, "Predictor")
        gui.comboBox(
            box, self, value='x_var', model=self.var_model, callback=self.apply)
        gui.spin(
            box, self, "polynomialexpansion", label="Polynomial degree: ",
            minv=0, maxv=10, alignment=Qt.AlignmentFlag.AlignRight,
            callback=self.apply)
        gui.checkBox(
            box, self, "fit_intercept",
            label="Fit intercept", callback=self.apply, stateWhenDisabled=True,
            tooltip="Add an intercept term;\n"
                    "This is always checked if the model is defined on input.")

        box = gui.vBox(self.controlArea, "Target")
        gui.comboBox(
            box, self, value="y_var", model=self.var_model, callback=self.apply)
        gui.checkBox(
            widget=box, master=self, value='error_bars_enabled',
            label="Show error bars", callback=self.apply)

        gui.rubber(self.controlArea)

        info_box = gui.vBox(self.controlArea, "Info")
        gui.label(
            widget=info_box, master=self,
            label="Regressor: %(regressor_name).30s")
        gui.label(
            widget=info_box, master=self,
            label="Mean absolute error: %(mae).6s")
        gui.label(
            widget=info_box, master=self,
            label="Root mean square error: %(rmse).6s")

        self.plotview = pg.PlotWidget(background="w")
        self.plot = self.plotview.getPlotItem()
        axis_pen = QPen(self.palette().color(QPalette.Text))
        tickfont = QFont(self.font())
        tickfont.setPixelSize(max(int(tickfont.pixelSize() * 2 // 3), 11))
        for axis in ("bottom", "left"):
            axis = self.plot.getAxis(axis)
            axis.setPen(axis_pen)
            axis.setTickFont(tickfont)
        self.plot.setRange(xRange=(0.0, 1.0), yRange=(0.0, 1.0),
                           disableAutoRange=True)
        self.mainArea.layout().addWidget(self.plotview)

    def add_bottom_buttons(self):
        pass

    def clear(self):
        self.data = None
        self.rmse = ""
        self.mae = ""
        self.clear_plot()

    def clear_plot(self):
        if self.plot_item is not None:
            self.plot_item.setParentItem(None)
            self.plotview.removeItem(self.plot_item)
            self.plot_item = None

        if self.scatterplot_item is not None:
            self.scatterplot_item.setParentItem(None)
            self.plotview.removeItem(self.scatterplot_item)
            self.scatterplot_item = None

        self.remove_error_items()
        self.plotview.clear()

    @check_sql_input
    def set_data(self, data):
        self.clear()
        self.Error.clear()
        self.closeContext()
        self.data = data
        if data is None:
            self.var_model.set_domain(None)
            return

        self.var_model.set_domain(data.domain)
        if len(self.var_model) < 2:
            self.data = None
            self.x_var = self.y_var = None
            self.Error.no_cont_variables()
            return

        self.x_var = self.var_model[0]
        if data.domain.class_var in self.var_model:
            self.y_var = data.domain.class_var
        else:
            self.y_var = self.var_model[1]
        self.openContext(data)

    @Inputs.learner
    def set_learner(self, learner):
        self.learner = learner
        if learner is None:
            self.controls.fit_intercept.setDisabled(False)
            self.regressor_name = self.default_learner_name
        else:
            self.controls.fit_intercept.setDisabled(True)
            self.regressor_name = learner.name

    def handleNewSignals(self):
        self.apply()

    def plot_scatter_points(self, x_data, y_data):
        if self.scatterplot_item:
            self.plotview.removeItem(self.scatterplot_item)
        self.scatterplot_item = pg.ScatterPlotItem(
            x=x_data, y=y_data,
            symbol="o", size=10, pen=pg.mkPen(0.2), brush=pg.mkBrush(0.7),
            antialias=True)
        self.scatterplot_item.opts["useCache"] = False
        self.plotview.addItem(self.scatterplot_item)
        self.plotview.replot()

    def set_range(self, x_data, y_data):
        min_x, max_x = np.nanmin(x_data), np.nanmax(x_data)
        min_y, max_y = np.nanmin(y_data), np.nanmax(y_data)
        if self.polynomialexpansion == 0 and not self._has_intercept:
            if min_y > 0:
                min_y = -0.1 * max_y
            elif max_y < 0:
                max_y = -0.1 * min_y

        self.plotview.setRange(
            QRectF(min_x, min_y, max_x - min_x, max_y - min_y),
            padding=0.025)
        self.plotview.replot()

    def plot_regression_line(self, x_data, y_data):
        item = pg.PlotCurveItem(
            x=x_data, y=y_data,
            pen=pg.mkPen(QColor(255, 0, 0), width=3),
            antialias=True
        )
        self._plot_regression_item(item)

    def plot_infinite_line(self, x, y, angle):
        item = pg.InfiniteLine(
            QPointF(x, y), angle,
            pen=pg.mkPen(QColor(255, 0, 0), width=3))
        self._plot_regression_item(item)

    def _plot_regression_item(self, item):
        if self.plot_item:
            self.plotview.removeItem(self.plot_item)
        self.plot_item = item
        self.plotview.addItem(self.plot_item)
        self.plotview.replot()

    def remove_error_items(self):
        for it in self.error_plot_items:
            self.plotview.removeItem(it)
        self.error_plot_items = []

    def plot_error_bars(self, x,  actual, predicted):
        self.remove_error_items()
        if self.error_bars_enabled:
            for x, a, p in zip(x, actual, predicted):
                line = pg.PlotCurveItem(
                    x=[x, x], y=[a, p],
                    pen=pg.mkPen(QColor(150, 150, 150), width=1),
                    antialias=True)
                self.plotview.addItem(line)
                self.error_plot_items.append(line)
        self.plotview.replot()

    def _varnames(self, name):
        # If variable name is short, use superscripts
        # otherwise "^" because superscripts would be lost
        def ss(x):
            # Compose a (potentially non-single-digit) superscript
            return "".join("⁰¹²³⁴⁵⁶⁷⁸⁹"[i] for i in (int(c) for c in str(x)))

        if len(name) <= 3:
            return [f"{name}{ss(i)}"
                    for i in range(not self._has_intercept,
                                   1 + self.polynomialexpansion)]
        else:
            return ["intercept"] * self._has_intercept + \
                   [name] * (self.polynomialexpansion >= 1) + \
                   [f"{name}^{i}" for i in range(2, 1 + self.polynomialexpansion)]

    @property
    def _has_intercept(self):
        return self.learner is not None or self.fit_intercept

    def apply(self):
        def error_and_clear(error=None):
            if error:
                error()
            self.clear_plot()
            self.Outputs.data.send(None)
            self.Outputs.coefficients.send(None)
            self.Outputs.learner.send(None)
            self.Outputs.model.send(None)

        self.Error.all_none.clear()
        self.Error.same_dep_indepvar.clear()
        self.Warning.large_diffs.clear()
        if self.data is None:
            error_and_clear()
            return
        if self.x_var is self.y_var:
            error_and_clear(self.Error.same_dep_indepvar)
            return

        degree = self.polynomialexpansion
        if degree == 0 and not self.fit_intercept and (
                self.learner is None
                or not getattr(self.learner, "fit_intercept", True)):
            learner = RegressTo0()
        else:
            # For LinearRegressionLearner, set fit_intercept to False:
            # the intercept is added as bias term in polynomial expansion
            learner = self.learner \
                      or LinearRegressionLearner(fit_intercept=False)

        include_bias = self.learner is None and self.fit_intercept
        poly_learner = PolynomialLearnerWrapper(
            self.x_var, self.y_var, degree, learner,
            self.preprocessors, include_bias)
        poly_learner.name = self.learner_name

        data_table, preprocessed_table, poly_preprocessor, \
        expanded_data, predictor = poly_learner.data_and_model(self.data)
        if preprocessed_table is None:
            error_and_clear(self.Error.all_none)
            return
        if expanded_data.X.max() - expanded_data.X.min() > 1e14:
            # the threshold defined with experimenting, instability typically
            # started to have effects for values >= 1e15
            self.Warning.large_diffs()

        model = None
        if hasattr(predictor, "model"):
            model = predictor.model
        elif hasattr(predictor, "skl_model"):
            model = predictor.skl_model

        x = preprocessed_table.X.ravel()
        y = preprocessed_table.Y.ravel()

        linspace = Table.from_numpy(
            Domain(data_table.domain.attributes),
            np.linspace(np.nanmin(x), np.nanmax(x), 1000).reshape(-1, 1),
        )
        values = predictor(linspace, predictor.Value)

        predicted = predictor(data_table, predictor.Value)
        results = Results(
            domain=self.data.domain,
            nrows=len(data_table), learners=[poly_learner],
            row_indices=np.arange(len(data_table)),
            folds=(Ellipsis,),
            actual=data_table.Y,
            predicted=predicted[None, :])
        self.rmse = round(RMSE(results)[0], 6)
        self.mae = round(MAE(results)[0], 6)

        # plot error bars
        self.plot_error_bars(x, results.actual, results.predicted.ravel())

        # plot data points
        self.plot_scatter_points(x, y)

        # plot regression line
        x_data, y_data = linspace.X.ravel(), values.ravel()
        if self.polynomialexpansion == 0:
            self.plot_infinite_line(x_data[0], y_data[0], 0)
        elif self.polynomialexpansion == 1 and self.learner is None:
            k = model.coef_[1 if self._has_intercept else 0]
            self.plot_infinite_line(x_data[0], y_data[0],
                                    math.degrees(math.atan(k)))
        else:
            self.plot_regression_line(x_data, y_data)

        self.plot.getAxis("bottom").setLabel(self.x_var.name)
        self.plot.getAxis("left").setLabel(self.y_var.name)
        self.set_range(x, y)

        self.Outputs.learner.send(poly_learner)
        self.Outputs.model.send(predictor)

        # Send model coefficents
        if model is not None and hasattr(model, "coef_"):
            if getattr(learner, "fit_intercept", True):
                coefs = [model.intercept_]
            else:
                coefs = []
            coefs += list(model.coef_)
        elif self.learner is None \
                and isinstance(predictor, MeanModel) \
                and self.fit_intercept:
            coefs = [predictor.mean]
        else:
            coefs = None
        if coefs:
            domain = Domain([ContinuousVariable("coef")],
                            metas=[StringVariable("name")])
            names = self._varnames(self.x_var.name)
            coef_table = Table.from_list(domain, list(zip(coefs, names)))
            self.Outputs.coefficients.send(coef_table)
        else:
            self.Outputs.coefficients.send(None)

        self.Outputs.data.send(expanded_data)

    def send_report(self):
        if self.data is None:
            return
        caption = report.render_items_vert((
            ("Polynomial Expansion", self.polynomialexpansion),
            ("Fit intercept",
             self._has_intercept and ["No", "Yes"][self.fit_intercept])
        ))
        self.report_plot()
        if caption:
            self.report_caption(caption)

    @classmethod
    def migrate_settings(cls, settings, version):
        # polynomialexpansion used to be controlled by doublespin and was hence
        # float. Just convert to `int`, ignore settings versions.
        settings["polynomialexpansion"] = \
            int(settings.get("polynomialexpansion", 1))


if __name__ == "__main__":
    learner = RidgeRegressionLearner(alpha=1.0)
    iris = Table("iris")
    WidgetPreview(OWPolynomialRegression).run(set_data=iris)  # , set_learner=learner)



import copy
from typing import Optional

import numpy as np
from AnyQt.QtCore import Qt, QAbstractTableModel, QModelIndex, QSize
from AnyQt.QtWidgets import QTableView, QItemDelegate, QLineEdit, QCompleter

from Orange.data import (
    Table,
    Domain,
    ContinuousVariable,
    DiscreteVariable,
    TimeVariable,
)
from Orange.widgets import gui
from Orange.widgets.settings import (
    Setting,
    ContextSetting,
    PerfectDomainContextHandler,
    ContextHandler,
)
from Orange.widgets.utils import vartype
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.widget import OWWidget, Output, Input, Msg


DEFAULT_DATA = [[None] * 3 for y in range(3)]


class EditableTableItemDelegate(QItemDelegate):
    def createEditor(self, parent, options, index: QModelIndex):
        model = index.model()  # type: EditableTableModel
        if not model.is_discrete(index.column()):
            return super().createEditor(parent, options, index)

        vals = model.discrete_vals(index.column())
        edit = QLineEdit(parent)
        edit.setCompleter(QCompleter(vals, edit, filterMode=Qt.MatchContains))

        def save():
            if edit.text():
                model.setData(index, edit.text())

        edit.editingFinished.connect(save)
        return edit

    def setEditorData(self, editor, index):
        current_val = index.model().data(index, Qt.DisplayRole)
        editor.setText(current_val)


class EditableTableModel(QAbstractTableModel):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._table = [[None]]
        self._domain = None

    def _var(self, index):
        # todo: remove if/when domain[idx] is fixed
        return (self._domain.variables + self._domain.metas)[index]

    def set_domain(self, domain: Optional[Domain]):
        self._domain = domain
        if domain is not None:
            # Todo: change to len(domain) when Domain len's behaviour is fixed
            n_columns = len(domain.variables) + len(domain.metas)
        else:
            n_columns = 3
        self.setColumnCount(n_columns)
        self.clear()

    def is_discrete(self, column):
        column_data = set(row[column] for row in self._table) - {None}

        def is_number(x):
            """
            Check if x is number
            x.is_digit only works for usigned ints this on works for ints and
            floats
            """
            try:
                float(x)
                return True
            except ValueError:
                return False

        return (
            self._domain is not None
            and self._var(column).is_discrete
            or (
                column_data
                and not self.is_time_variable(column)
                and not all(map(lambda s: is_number(s), column_data))
            )
        )

    def is_time_variable(self, column):
        values = self.time_vals(column)
        return values and not all(e is None for e in values)

    def time_vals(self, column):
        column_data = [row[column] for row in self._table]
        try:
            tvar = TimeVariable("_")
            values = [
                tvar.parse_exact_iso(d) if d is not None else None
                for d in column_data
            ]
            return values
        except ValueError:
            return None

    def discrete_vals(self, column):
        if self._domain is not None and self._var(column).is_discrete:
            return self._var(column).values
        else:
            return list(set(row[column] for row in self._table) - {None})

    def rowCount(self, parent=None):
        return len(self._table)

    def columnCount(self, parent=None):
        return len(self._table[0])

    def flags(self, index):
        return Qt.ItemIsEnabled | Qt.ItemIsEditable

    def data(self, index: QModelIndex, role=None):
        value = self._table[index.row()][index.column()]
        if role == Qt.DisplayRole and value is not None:
            return str(value)
        elif role == Qt.TextAlignmentRole:
            return Qt.AlignVCenter | Qt.AlignLeft

    def setData(self, index: QModelIndex, value: str, role=None):
        value = None if not value else value
        self._table[index.row()][index.column()] = value
        self.dataChanged.emit(index, index)
        return True

    def set_table(self, table):
        self.beginResetModel()
        # must be a copy since model changes it inplace
        self._table = copy.deepcopy(table)
        self.endResetModel()

    def get_raw_table(self):
        # the model updates the table inplace
        return copy.deepcopy(self._table)

    def setRowCount(self, n_rows):
        diff = n_rows - self.rowCount()
        if diff > 0:
            self.beginInsertRows(QModelIndex(), self.rowCount(), n_rows - 1)
            self._table += [[None] * self.columnCount() for _ in range(diff)]
            self.endInsertRows()
        elif diff < 0:
            self.beginRemoveRows(QModelIndex(), n_rows, self.rowCount() - 1)
            del self._table[n_rows:]
            self.endRemoveRows()

    def setColumnCount(self, n_columns):
        diff = n_columns - self.columnCount()
        if diff > 0:
            self.beginInsertColumns(
                QModelIndex(), self.columnCount(), n_columns - 1)
            for row in self._table:
                row += [None] * diff
            self.endInsertColumns()
        elif diff < 0:
            self.beginRemoveColumns(
                QModelIndex(), n_columns, self.columnCount() - 1)
            for row in self._table:
                del row[n_columns:]
            self.endRemoveColumns()

    def headerData(self, section, orientation, role=None):
        if orientation == Qt.Vertical:
            return super().headerData(section, orientation, role)

        if role == Qt.DisplayRole:
            if self._domain is None:
                return str(section + 1)
            else:
                return self._var(section).name

    def clear(self):
        self.set_table(
            [[None] * self.columnCount()
             for _ in range(self.rowCount())]
        )

    def get_table(self):
        domain = self.get_domain()
        data = np.array(self._table)  # type:
        for ci in range(data.shape[1]):
            if isinstance((domain.variables + domain.metas)[ci], TimeVariable):
                data[:, ci] = self.time_vals(ci)
        return Table.from_list(domain, data)

    def get_domain(self):
        if self._domain is not None:
            return self._domain

        vars = []
        for ci in range(self.columnCount()):
            if self.is_discrete(ci):
                values = set(
                    row[ci] for row in self._table if row[ci] is not None
                )
                var = DiscreteVariable(name=str(ci + 1), values=values)
            elif self.is_time_variable(ci):
                var = TimeVariable(name=str(ci + 1))
            else:
                var = ContinuousVariable(name=str(ci + 1))
            vars.append(var)
        return Domain(vars)


class CreateTableContextHandler(PerfectDomainContextHandler):
    def open_context(self, widget, domain):
        ContextHandler.open_context(
            self, widget, domain, *self.encode_domain(domain)
        )

    def encode_domain(self, domain):
        """
        Encode domain into tuples (name, type)
        A tuple is returned for each of attributes, class_vars and metas.
        """
        if self.match_values == self.MATCH_VALUES_ALL:

            def _encode(attrs):
                return tuple(
                    (v.name, list(v.values) if v.is_discrete else vartype(v))
                    for v in attrs
                )

        else:

            def _encode(attrs):
                return tuple((v.name, vartype(v)) for v in attrs)

        if domain is None:
            return (None, None, None)
        return (
            _encode(domain.attributes),
            _encode(domain.class_vars),
            _encode(domain.metas),
        )


class OWCreateTable(OWWidget):
    name = "Create Table"
    icon = "icons/CreateTable.svg"
    priority = 50
    keywords = []

    class Inputs:
        data = Input("Data", Table)

    class Outputs:
        data = Output("Data", Table)

    class Error(OWWidget.Error):
        transform_err = Msg("Data does not fit to domain")

    settingsHandler = CreateTableContextHandler()

    n_rows = Setting(len(DEFAULT_DATA))
    n_columns = Setting(len(DEFAULT_DATA[0]))
    auto_commit = Setting(True)
    # since data is a small (at most 20x20) table we can afford to store it
    # as a context
    context_data = ContextSetting(copy.deepcopy(DEFAULT_DATA), schema_only=True)

    def __init__(self):
        super().__init__()

        options = {"labelWidth": 100, "controlWidth": 50}
        box = gui.vBox(self.controlArea, "Control")
        self.r_spin = gui.spin(
            box,
            self,
            "n_rows",
            1,
            20,
            1,
            **options,
            label="Rows:",
            callback=self.nrows_changed
        )
        self.c_spin = gui.spin(
            box,
            self,
            "n_columns",
            1,
            20,
            1,
            **options,
            label="Columns:",
            callback=self.ncolumns_changed
        )
        box.setMinimumWidth(200)

        gui.rubber(self.controlArea)
        gui.auto_send(self.buttonsArea, self, "auto_commit")

        box = gui.vBox(self.mainArea, True, margin=0)
        self.table = QTableView(box)
        self.table.setItemDelegate(EditableTableItemDelegate())
        self.table.setEditTriggers(self.table.CurrentChanged)
        box.layout().addWidget(self.table)

        self.table_model = EditableTableModel()
        self.table.setModel(self.table_model)
        self.table_model.dataChanged.connect(self.data_changed)
        self.table_model.set_table(self.context_data)
        self.set_dataset(None)  # to init the context

    def nrows_changed(self):
        self.table_model.setRowCount(self.n_rows)
        self.commit()

    def ncolumns_changed(self):
        self.table_model.setColumnCount(self.n_columns)
        self.commit()

    def data_changed(self):
        self.context_data = self.table_model.get_raw_table()
        self.commit()

    def commit(self):
        data = None
        try:
            data = self.table_model.get_table()
            self.Error.transform_err.clear()
        except Exception as ex:
            self.Error.transform_err()
        self.Outputs.data.send(data)

    @Inputs.data
    def set_dataset(self, data):
        self.closeContext()
        if data is not None:
            self.table_model.set_domain(data.domain)
            self.context_data = [
                [None] * (len(data.domain.variables) + len(data.domain.metas))
                for _ in range(self.table_model.rowCount())
            ]
        else:
            self.table_model.set_domain(None)
            self.context_data = copy.deepcopy(DEFAULT_DATA)
        self.c_spin.setEnabled(data is None)
        self.c_spin.setValue(self.table_model.columnCount())
        self.openContext(data)
        self.unconditional_commit()

    @staticmethod
    def sizeHint():
        return QSize(800, 500)

    def openContext(self, data):
        super(OWCreateTable, self).openContext(data.domain if data else None)
        self.table_model.set_table(self.context_data)


if __name__ == "__main__":  # pragma: no cover
    WidgetPreview(OWCreateTable).run(Table("iris"))



import unittest

from orangecontrib.educational.widgets.utils.kmeans import Kmeans
import numpy as np


class TestKmeans(unittest.TestCase):
    def setUp(self):
        self.data = np.arange(20).reshape(10, 2)
        self.kmeans = Kmeans(self.data)

    def test_converged(self):
        """It doesn't converge in a half-step, but does eventually"""
        self.kmeans.move_centroid(0, 0, 1)
        self.kmeans.move_centroid(1, 2, 3)
        self.kmeans.move_centroid(2, 4, 5)
        i = -1
        for i in range(50):
            self.kmeans.move_centroids()
            self.kmeans.assign_membership()
            if self.kmeans.converged:
                break
        self.assertGreater(i, 1)
        self.assertTrue(self.kmeans.converged)

    def test_reset(self):
        assert self.kmeans.k == len(self.kmeans.centroids) == 3
        centroids = self.kmeans.centroids.copy()
        self.kmeans.reset(self.data[:9])
        np.testing.assert_equal(self.kmeans.data, self.data[:9])
        self.assertFalse(np.all(centroids == self.kmeans.centroids.copy()))

        self.kmeans.reset(self.data, 5)
        np.testing.assert_equal(self.kmeans.data, self.data)
        self.assertEqual(self.kmeans.k, 5)

    def test_find_clusters(self):
        # if there is only one cluster, all points belong to it
        self.kmeans.reset(self.data, 1)
        np.testing.assert_equal(self.kmeans._find_clusters(), 0)

        # Have two centroids that split the data
        self.kmeans.reset(self.data, 2)
        self.kmeans.move_centroid(1, 4, 5)
        self.kmeans.move_centroid(0, 6, 7)
        np.testing.assert_equal(self.kmeans._find_clusters(), [1] * 3 + [0] * 7)

    def test_steps(self):
        self.kmeans.reset(self.data, 2)
        self.kmeans.centroids = np.array([[6, 7], [4, 5]])
        self.kmeans.assign_membership()
        np.testing.assert_equal(self.kmeans.clusters, [1] * 3 + [0] * 7)

        self.kmeans.move_centroids()
        np.testing.assert_equal(self.kmeans.centroids, [[12, 13], [2, 3]])

    def test_history(self):
        assert not self.kmeans.history

        centroids = self.kmeans.centroids.copy()
        self.kmeans.move_centroids()
        self.assertIs(self.kmeans.history[0].step, Kmeans.move_centroids)

        clusters = self.kmeans.clusters.copy()
        self.kmeans.assign_membership()
        self.assertIs(self.kmeans.history[1].step, Kmeans.assign_membership)

        self.assertEqual(len(self.kmeans.history), 2)

        self.kmeans.step_back()
        self.assertEqual(len(self.kmeans.history), 1)
        self.assertIs(self.kmeans.history[0].step, Kmeans.move_centroids)
        np.testing.assert_equal(self.kmeans.clusters, clusters)

        self.kmeans.step_back()
        self.assertEqual(len(self.kmeans.history), 0)
        np.testing.assert_equal(self.kmeans.centroids, centroids)

    def test_add_centroid(self):
        centroids = self.kmeans.centroids.tolist()
        assert len(centroids) == 3
        self.kmeans.add_centroid(4, 5)
        self.assertEqual(self.kmeans.k, 4)
        self.assertEqual(self.kmeans.centroids.tolist(), centroids + [[4, 5]])
        self.assertIs(self.kmeans.history[-1].step, Kmeans.add_centroid)

        self.kmeans.step_back()
        self.assertEqual(self.kmeans.centroids.tolist(), centroids)

    def test_delete_centroid(self):
        centroids = self.kmeans.centroids.tolist()
        assert len(centroids) == 3
        self.kmeans.delete_centroid(1)
        self.assertEqual(self.kmeans.k, 2)
        self.assertEqual(self.kmeans.centroids.tolist(),
                         centroids[:1] + centroids[2:])
        self.assertIs(self.kmeans.history[-1].step, Kmeans.delete_centroid)

        self.kmeans.step_back()
        self.assertEqual(self.kmeans.centroids.tolist(), centroids)

    def test_move_centroid(self):
        c0, c1, c2 = self.kmeans.centroids.tolist()
        cn = [3, 3.2]
        self.kmeans.move_centroid(1, *cn)
        self.assertEqual(self.kmeans.centroids.tolist(), [c0, cn, c2])
        self.assertEqual(self.kmeans.k, 3)
        self.assertIs(self.kmeans.history[-1].step, Kmeans.move_centroid)

        self.kmeans.step_back()
        self.assertEqual(self.kmeans.centroids.tolist(), [c0, c1, c2])


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

from Orange.data import Table, Domain, ContinuousVariable
from Orange.widgets.tests.base import WidgetTest
from orangecontrib.educational.widgets.owpolynomialregression \
    import OWPolynomialRegression, PolynomialFeatures, RegressTo0, \
    TempMeanModel, PolynomialLearnerWrapper
from Orange.regression import (LinearRegressionLearner,
                               RandomForestRegressionLearner)
from Orange.regression.tree import TreeLearner as TreeRegressionLearner
from Orange.preprocess.preprocess import Normalize


class TestOWPolynomialRegression(WidgetTest):
    def setUp(self):
        self.widget = self.create_widget(OWPolynomialRegression)   # type: OWPolynomialRegression
        self.data = Table.from_file("iris")
        self.data_housing = Table.from_file("housing")

    def test_set_data(self):
        variables = self.data.domain.variables
        class_variables = self.data.domain.class_vars
        continuous_variables = [var for var in variables if var.is_continuous]

        self.widget.set_data(self.data)

        self.assertEqual(self.data, self.widget.data)
        self.assertEqual(continuous_variables, list(self.widget.var_model))
        self.assertEqual(self.widget.x_var, continuous_variables[0])
        self.assertEqual(self.widget.y_var, continuous_variables[1])

        # check for none data
        self.widget.set_data(None)
        self.assertEqual(self.widget.data, None)

        # data set with continuous class to check if nclass > 0
        variables = self.data_housing.domain.variables
        continuous_variables = [var for var in variables if var.is_continuous]

        self.widget.set_data(self.data_housing)
        self.assertEqual(self.data_housing, self.widget.data)
        self.assertEqual(continuous_variables, list(self.widget.var_model))
        self.assertEqual(self.widget.x_var, continuous_variables[0])
        self.assertEqual(self.widget.y_var, self.data_housing.domain.class_var)

        # check with data with all none
        data = Table.from_list(
            Domain([ContinuousVariable('a'),
                    ContinuousVariable('b')]),
            [[None, None], [None, None]])
        self.widget.set_data(data)
        self.widget.apply()
        self.assertIsNone(self.widget.plot_item)
        self.assertIsNone(self.widget.scatterplot_item)

    def test_add_main_layout(self):
        w = self.widget
        self.assertEqual(w.data, None)
        self.assertEqual(w.preprocessors, None)
        self.assertEqual(w.learner, None)
        self.assertEqual(w.scatterplot_item, None)
        self.assertEqual(w.plot_item, None)

        self.assertEqual(
            w.controls.regressor_name.text(), "Regressor: Linear Regression")

    def test_send_report(self):
        # check if nothing happens when polynomialexpansion is None
        self.widget.polynomialexpansion = None
        self.widget.send_report()
        self.assertNotIn("class='caption'", self.widget.report_html)
        self.widget.polynomialexpansion = 1

        self.widget.send_report()
        self.assertEqual(self.widget.report_html, "")

        self.widget.set_data(self.data)
        self.widget.send_report()

        self.assertNotEqual(self.widget.report_html, "")

    def test_clear(self):
        self.widget.set_data(self.data)
        self.widget.clear()

        self.assertEqual(self.widget.data, None)
        #just check if clear function also call clear_plot
        self.assertEqual(self.widget.plot_item, None)
        self.assertEqual(self.widget.scatterplot_item, None)
        self.assertEqual(len(self.widget.error_plot_items), 0)

    def test_clear_plot(self):
        self.widget.set_data(self.data)
        self.widget.apply()
        self.widget.clear_plot()

        self.assertEqual(self.widget.plot_item, None)
        self.assertEqual(self.widget.scatterplot_item, None)

    def test_set_learner(self):
        w = self.widget

        lin = LinearRegressionLearner
        lin.name = "Linear Regression"
        self.widget.set_learner(lin)
        self.assertEqual(self.widget.learner, lin)

        self.assertEqual(
            w.controls.regressor_name.text(), "Regressor: Linear Regression")

        tree = TreeRegressionLearner
        tree.name = "Tree Learner"

        self.widget.set_learner(tree)
        self.assertEqual(self.widget.learner, tree)
        self.assertEqual(
            w.controls.regressor_name.text(), "Regressor: Tree Learner")

    def test_plot_scatter_points(self):
        x_data = [1, 2, 3]
        y_data = [2, 3, 4]

        self.widget.plot_scatter_points(x_data, y_data)

        self.assertNotEqual(self.widget.scatterplot_item, None)

        # check case when scatter plot allready exist
        x_data = [2, 3, 5]
        y_data = [2, 3, 10]

        self.widget.plot_scatter_points(x_data, y_data)

        self.assertNotEqual(self.widget.scatterplot_item, None)

    def test_plot_regression_line(self):
        x_data = [1, 2, 3]
        y_data = [2, 3, 4]

        self.widget.plot_regression_line(x_data, y_data)

        self.assertNotEqual(self.widget.plot_item, None)

        # check case when regression_line already exist
        x_data = [2, 3, 5]
        y_data = [2, 3, 10]

        self.widget.plot_regression_line(x_data, y_data)

        self.assertNotEqual(self.widget.plot_item, None)

    def test_plot_error_bars(self):
        w = self.widget

        check = w.controls.error_bars_enabled
        check.click()

        x_data = [1, 2, 3]
        y_data = [2, 3, 4]
        y_data_fake = [1, 2, 4]

        self.widget.plot_error_bars(x_data, y_data, y_data_fake)
        self.assertEqual(len(w.error_plot_items), len(x_data))

        check.click()

        self.widget.plot_error_bars(x_data, y_data, y_data_fake)
        self.assertEqual(len(w.error_plot_items), 0)

        check.click()

        self.send_signal(w.Inputs.data, self.data)
        self.assertEqual(len(w.error_plot_items), len(self.data))

    def test_apply(self):
        self.widget.set_data(self.data)
        self.widget.apply()

        self.assertNotEqual(self.widget.plot_item, None)
        self.assertNotEqual(self.widget.scatterplot_item, None)

        self.widget.set_data(None)
        self.widget.apply()
        # TODO: output will be checked when it available in GuiTest

        # check if function does not change plots that are None according to test_set_data
        self.assertEqual(self.widget.plot_item, None)
        self.assertEqual(self.widget.scatterplot_item, None)

        self.widget.set_data(self.data)
        self.widget.set_learner(LinearRegressionLearner())
        self.widget.apply()

        self.assertNotEqual(self.widget.plot_item, None)
        self.assertNotEqual(self.widget.scatterplot_item, None)

        self.widget.set_learner(RandomForestRegressionLearner())
        self.widget.apply()

        self.assertNotEqual(self.widget.plot_item, None)
        self.assertNotEqual(self.widget.scatterplot_item, None)

        self.widget.set_preprocessor(Normalize())
        self.assertNotEqual(self.widget.plot_item, None)
        self.assertNotEqual(self.widget.scatterplot_item, None)

    def test_data_output(self):
        """
        Check if correct data on output
        """
        w = self.widget
        spin =w.controls.polynomialexpansion

        self.assertIsNone(self.get_output(w.Outputs.data))

        u, x, y, z = (ContinuousVariable(n) for n in "uxyz")
        domain = Domain([u, x], y)
        data = Table.from_numpy(
            domain,
            [[1, 1], [0, 2], [np.nan, 3], [-1, np.nan], [2, 4]],
            [3, 5, 7, 7, np.nan])

        spin.setValue(0)
        self.send_signal(w.Inputs.data, data)
        w.x_var = x
        w.y_var = y

        w.fit_intercept = False
        spin.setValue(1)
        np.testing.assert_almost_equal(
            self.get_output(w.Outputs.data).X.T,
            [[1, 2, 3]])

        spin.setValue(2)
        np.testing.assert_almost_equal(
            self.get_output(w.Outputs.data).X.T,
            [[1, 2, 3], [1, 4, 9]])

        spin.setValue(3)
        np.testing.assert_almost_equal(
            self.get_output(w.Outputs.data).X.T,
            [[1, 2, 3], [1, 4, 9], [1, 8, 27]])

        w.fit_intercept = True
        spin.setValue(1)
        np.testing.assert_almost_equal(
            self.get_output(w.Outputs.data).X.T,
            [[1, 1, 1], [1, 2, 3]])

        spin.setValue(2)
        np.testing.assert_almost_equal(
            self.get_output(w.Outputs.data).X.T,
            [[1, 1, 1], [1, 2, 3], [1, 4, 9]])

        spin.setValue(3)
        np.testing.assert_almost_equal(
            self.get_output(w.Outputs.data).X.T,
            [[1, 1, 1], [1, 2, 3], [1, 4, 9], [1, 8, 27]])

        self.send_signal(w.Inputs.data, None)
        self.assertIsNone(self.get_output(w.Outputs.data))

    def test_data_nan_row(self):
        """
        When some rows are nan in attributes array widget crashes.
        GH-43
        """
        data = self.data[::50].copy()
        with data.unlocked(data.X):
            data.X[0] = np.nan
        self.send_signal(self.widget.Inputs.data, data)

    def test_coefficients(self):
        w = self.widget

        u, x, y, z = (ContinuousVariable(n) for n in "uxyz")
        domain = Domain([u, x], y)
        data = Table.from_numpy(
            domain,
            [[1, 1], [0, 2], [np.nan, 3], [-1, np.nan], [2, 4]],
            [3, 5, 7, 7, np.nan])

        self.send_signal(w.Inputs.data, data)
        w.x_var = x
        w.y_var = y

        spin = self.widget.controls.polynomialexpansion
        intercept_cb = self.widget.controls.fit_intercept
        intercept_cb.setChecked(True)

        spin.setValue(0)
        coef = self.get_output(w.Outputs.coefficients)
        np.testing.assert_almost_equal(coef.X, [[5]])

        spin.setValue(1)
        coef = self.get_output(w.Outputs.coefficients)
        np.testing.assert_almost_equal(coef.X.T, [[1, 2]])

        spin.setValue(2)
        coef = self.get_output(w.Outputs.coefficients)
        np.testing.assert_almost_equal(coef.X.T, [[1, 2, 0]])

        intercept_cb.setChecked(False)
        spin.setValue(0)
        coef = self.get_output(w.Outputs.coefficients)
        self.assertIsNone(coef)

        spin.setValue(1)
        coef = self.get_output(w.Outputs.coefficients)
        # I haven't computed this value manually, I just copied it
        np.testing.assert_almost_equal(coef.X.T, [[2.4285714]])

        spin.setValue(2)
        coef = self.get_output(w.Outputs.coefficients)
        # I haven't computed these values manually, I just copied them
        np.testing.assert_almost_equal(coef.X.T, [[ 3.1052632, -0.2631579]])

        self.send_signal(w.Inputs.learner, LinearRegressionLearner(fit_intercept=True))

        spin.setValue(0)
        coef = self.get_output(w.Outputs.coefficients)
        np.testing.assert_almost_equal(coef.X.T, [[5]])

        spin.setValue(1)
        coef = self.get_output(w.Outputs.coefficients)
        np.testing.assert_almost_equal(coef.X.T, [[1, 2]])

        spin.setValue(2)
        coef = self.get_output(w.Outputs.coefficients)
        np.testing.assert_almost_equal(coef.X.T, [[1, 2, 0]])

        self.send_signal(w.Inputs.learner, LinearRegressionLearner(fit_intercept=False))

        spin.setValue(0)
        coef = self.get_output(w.Outputs.coefficients)
        self.assertIsNone(coef)

        spin.setValue(1)
        coef = self.get_output(w.Outputs.coefficients)
        # I haven't computed this value manually, I just copied it
        np.testing.assert_almost_equal(coef.X.T, [[2.4285714]])

        spin.setValue(2)
        coef = self.get_output(w.Outputs.coefficients)
        # I haven't computed these values manually, I just copied them
        np.testing.assert_almost_equal(coef.X.T, [[ 3.1052632, -0.2631579]])

    def test_large_polynomial_values(self):
        x, y = (ContinuousVariable(n) for n in "xy")
        data = Table.from_numpy(Domain([x], y), [[1], [30]], [3, 5])
        self.send_signal(self.widget.Inputs.data, data)

        self.widget.controls.polynomialexpansion.setValue(9)
        self.assertFalse(self.widget.Warning.large_diffs.is_shown())
        self.widget.controls.polynomialexpansion.setValue(10)
        self.assertTrue(self.widget.Warning.large_diffs.is_shown())
        self.widget.controls.polynomialexpansion.setValue(9)
        self.assertFalse(self.widget.Warning.large_diffs.is_shown())


class PolynomialFeaturesTest(unittest.TestCase):
    def test_1d(self):
        x, y, z = (ContinuousVariable(n) for n in "xyz")
        domain = Domain([x], y, [z])
        data = Table.from_numpy(
            domain,
            [[1], [2], [3]], [1, 2, 3], [[1], [2], [3]])
        data2 = Table.from_numpy(
            domain,
            [[3], [4], [5]], [1, 2, 3], [[1], [2], [3]])
        tf = PolynomialFeatures(1, False)(data)
        self.assertIs(tf.domain.class_var, y)
        np.testing.assert_equal(tf.Y, [1, 2, 3])
        self.assertEqual(tf.domain.metas, (z, ))
        np.testing.assert_equal(tf.metas.T, [[1, 2, 3]])

        np.testing.assert_equal(tf.X.T, [[1, 2, 3]])
        np.testing.assert_equal(data2.transform(tf.domain).X.T, [[3, 4, 5]])

        tf = PolynomialFeatures(1, True)(data)
        np.testing.assert_equal(
            tf.X.T,
            [[1, 1, 1], [1, 2, 3]])
        np.testing.assert_equal(
            data2.transform(tf.domain).X.T,
            [[1, 1, 1], [3, 4, 5]])

        tf = PolynomialFeatures(2, True)(data)
        np.testing.assert_equal(
            tf.X.T,
            [[1, 1, 1], [1, 2, 3], [1, 4, 9]])
        np.testing.assert_equal(
            data2.transform(tf.domain).X.T,
            [[1, 1, 1], [3, 4, 5], [9, 16, 25]])

        tf = PolynomialFeatures(2, False)(data)
        np.testing.assert_equal(
            tf.X.T,
            [[1, 2, 3], [1, 4, 9]])
        np.testing.assert_equal(
            data2.transform(tf.domain).X.T,
            [[3, 4, 5], [9, 16, 25]])

        tf = PolynomialFeatures(3, True)(data)
        np.testing.assert_equal(
            tf.X.T,
            [[1, 1, 1], [1, 2, 3], [1, 4, 9], [1, 8, 27]])
        np.testing.assert_equal(
            data2.transform(tf.domain).X.T,
            [[1, 1, 1], [3, 4, 5], [9, 16, 25], [27, 64, 125]])

        tf = PolynomialFeatures(3, False)(data)
        np.testing.assert_equal(
            tf.X.T,
            [[1, 2, 3], [1, 4, 9], [1, 8, 27]])
        np.testing.assert_equal(
            data2.transform(tf.domain).X.T,
            [[3, 4, 5], [9, 16, 25], [27, 64, 125]])

        tf = PolynomialFeatures(0, True)(data)
        np.testing.assert_equal(
            tf.X.T,
            [[1, 1, 1]])
        np.testing.assert_equal(
            data2.transform(tf.domain).X.T,
            [[1, 1, 1]])

        tf = PolynomialFeatures(0, False)(data)
        np.testing.assert_equal(
            tf.X.T,
            [[0, 0, 0]])
        np.testing.assert_equal(
            data2.transform(tf.domain).X.T,
            [[0, 0, 0]])

    def test_nd(self):
        x, y, z = (ContinuousVariable(n) for n in "xyz")
        domain = Domain([x, y, z])
        data = Table.from_numpy(
            domain,
            [[1, 2, 3], [4, 5, 6]])
        data2 = Table.from_numpy(
            domain,
            [[1, 3, 5]])

        tf = PolynomialFeatures(1, False)(data)
        np.testing.assert_equal(tf.X, data.X)
        np.testing.assert_equal(data2.transform(tf.domain).X, data2.X)

        tf = PolynomialFeatures(1, True)(data)
        np.testing.assert_equal(tf.X, [[1, 1, 2, 3], [1, 4, 5, 6]])
        np.testing.assert_equal(data2.transform(tf.domain).X, [[1, 1, 3, 5]])

        tf = PolynomialFeatures(2, False)(data)
        np.testing.assert_equal(
            tf.X,
            [[1, 2, 3, 1, 2, 3, 4, 6, 9],
             [4, 5, 6, 16, 20, 24, 25, 30, 36]])
        np.testing.assert_equal(
            data2.transform(tf.domain).X,
            [[1, 3, 5, 1, 3, 5, 9, 15, 25]])

        tf = PolynomialFeatures(2, True)(data)
        np.testing.assert_equal(
            tf.X,
            [[1, 1, 2, 3, 1, 2, 3, 4, 6, 9],
             [1, 4, 5, 6, 16, 20, 24, 25, 30, 36]])
        np.testing.assert_equal(
            data2.transform(tf.domain).X,
            [[1, 1, 3, 5, 1, 3, 5, 9, 15, 25]])

        tf = PolynomialFeatures(0, True)(data)
        np.testing.assert_equal(
            tf.X.T,
            [[1, 1]])
        np.testing.assert_equal(
            data2.transform(tf.domain).X.T,
            [[1]])

        tf = PolynomialFeatures(0, False)(data)
        np.testing.assert_equal(
            tf.X.T,
            [[0, 0]])
        np.testing.assert_equal(
            data2.transform(tf.domain).X.T,
            [[0]])


class ModelsTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.data = Table("iris")[:10]

    def test_regressto0(self):
        model = RegressTo0()(self.data)
        prediction = model(self.data[:5])
        np.testing.assert_equal(prediction, [0] * 5)

    def test_tempmeanmodel(self):
        model = TempMeanModel(42)
        prediction = model(self.data[:5])
        np.testing.assert_equal(prediction, [42] * 5)

    def test_polynomiallearnerwrapper(self):
        u, x, y, z = (ContinuousVariable(n) for n in "uxyz")
        domain = Domain([u, x], y)
        data = Table.from_numpy(
            domain,
            [[1, 1], [0, 2], [np.nan, 3], [-1, np.nan], [2, 4]],
            [3, 5, 7, 7, np.nan])
        data2 = Table.from_numpy(
            Domain([x, z]),
            [[5, 6], [7, 8]])

        learner = PolynomialLearnerWrapper(
            x, y, 1, LinearRegressionLearner(fit_intercept=False),
            fit_intercept=True, preprocessors=None)
        model = learner(data)
        np.testing.assert_almost_equal(model.coefficients, [1, 2])
        np.testing.assert_almost_equal(model(data2), [11, 15])

        learner = PolynomialLearnerWrapper(
            x, y, 1, LinearRegressionLearner(fit_intercept=True),
            fit_intercept=False, preprocessors=False)
        model = learner(data)
        np.testing.assert_almost_equal(model.coefficients, [2])
        self.assertAlmostEqual(model.intercept, 1)
        np.testing.assert_almost_equal(model(data2), [11, 15])

        learner = PolynomialLearnerWrapper(
            x, y, 1, LinearRegressionLearner(fit_intercept=False),
            fit_intercept=False, preprocessors=None)
        model = learner(data)
        # I haven't computed this value manually, just copied the result
        # But it must be something a bit larger than 2, hence...
        np.testing.assert_almost_equal(model.coefficients, [2.4285714])
        self.assertEqual(model.intercept, 0)
        np.testing.assert_almost_equal(model(data2), [12.1428571, 17])

        learner = PolynomialLearnerWrapper(
            x, y, 2, LinearRegressionLearner(fit_intercept=False),
            fit_intercept=False, preprocessors=False)
        model = learner(data)
        # I haven't computed this value manually, just copied the result
        np.testing.assert_almost_equal(model.coefficients,
                                       [3.1052632, -0.2631579])
        self.assertEqual(model.intercept, 0)
        np.testing.assert_almost_equal(model(data2), [8.9473684, 8.8421053])


if __name__ == "__main__":
    unittest.main()



import time
from colorsys import rgb_to_hsv, hsv_to_rgb

import numpy as np
from scipy.interpolate import splprep, splev

from AnyQt.QtCore import Qt, QObject, pyqtSignal, QThread, QEvent, QRectF
from AnyQt.QtWidgets import QGraphicsSceneMouseEvent, QGraphicsTextItem
from AnyQt.QtGui import QCursor, QColor

import pyqtgraph as pg

from Orange.classification import Model
from Orange.data import \
    ContinuousVariable, DiscreteVariable, StringVariable, Domain, Table
from Orange.preprocess.transformation import Indicator
from Orange.widgets import gui
from Orange.widgets import settings
from Orange.widgets.utils.itemmodels import VariableListModel
from Orange.widgets.widget import OWWidget, Msg, Input, Output
from Orange.preprocess.preprocess import Normalize
from orangewidget.report import report
from orangewidget.utils.widgetpreview import WidgetPreview

from orangecontrib.educational.widgets.utils.gradient_grid import \
    interpolate_grid
from orangecontrib.educational.widgets.utils.linear_regression import \
    LinearRegression
from orangecontrib.educational.widgets.utils.logistic_regression \
    import LogisticRegression
from orangecontrib.educational.widgets.utils.contour import Contour

GRID_SIZE = 20


class Autoplay(QThread):
    """
    Class used for separated thread when using "Autoplay" for gradient descent

    Parameters
    ----------
    ow_gradient_descent : OWGradientDescent
        Instance of OWGradientDescent class
    """

    def __init__(self, ow_gradient_descent):
        QThread.__init__(self)
        self.ow_gradient_descent = ow_gradient_descent

    def __del__(self):
        self.wait()

    def run(self):
        """
        Stepping through the algorithm until converge or user interrupts
        """
        while (self.ow_gradient_descent.learner and
               not self.ow_gradient_descent.learner.converged and
               self.ow_gradient_descent.auto_play_enabled and
               self.ow_gradient_descent.learner.step_no <= 500):
            try:
                self.ow_gradient_descent.step_trigger.emit()
            except RuntimeError:
                return
            time.sleep(0.2)
        self.ow_gradient_descent.stop_auto_play_trigger.emit()


class HoverEventDelegate(QObject):
    def __init__(self, delegate, parent=None):
        super().__init__(parent)
        self.delegate = delegate

    def eventFilter(self, _, event):
        return isinstance(event, QGraphicsSceneMouseEvent) \
               and self.delegate(event)


class OWGradientDescent(OWWidget):
    name = "Gradient Descent"
    description = "Demonstration of gradient descent " \
                  "in logistic or linear regression"
    keywords = ["gradient descent", "optimization", "gradient"]
    icon = "icons/GradientDescent.svg"
    want_main_area = True
    priority = 400

    class Inputs:
        data = Input("Data", Table)

    class Outputs:
        model = Output("Model", Model)
        coefficients = Output("Coefficients", Table)

    graph_name = "graph"  # QGraphicsView (pg.PlotWidget)

    attr_x_hint = settings.Setting("")
    attr_y_hint = settings.Setting("")
    normalization = settings.Setting(True)
    target_class_hint = settings.Setting('')
    alpha = settings.Setting(0.4)
    step_size = settings.Setting(30)  # step size for stochastic gds
    auto_play_speed = settings.Setting(1)
    stochastic = settings.Setting(False)

    default_background_color = [0, 0xbf, 0xff]
    auto_play_button_text = ["Run", "Stop"]

    step_trigger = pyqtSignal()
    stop_auto_play_trigger = pyqtSignal()

    class Error(OWWidget.Error):
        num_features = Msg("Data must contain at least {}.")
        no_class = Msg("Data must have a single target variable.")
        no_class_values = Msg("Target variable must have at least two values.")
        no_nonnan_data = Msg("No points with defined values.")
        same_variable = Msg("Select two different variables.")

    def __init__(self):
        super().__init__()
        self.learner = None
        self.data = None
        self.attr_x = self.attr_y = None
        self.target_class = ""
        self.selected_data = None
        self.cost_grid = None
        self.min_x = self.max_x = self.min_y = self.max_y = None
        self.contours = []
        self.optimization_path = self.last_point = None
        self.hover_label = self.optimization_label = None

        self.auto_play_enabled = False
        self.auto_play_button_text = ["Run", "Stop"]
        self.auto_play_thread = None

        self.var_model = VariableListModel()

        self.options_box = gui.widgetBox(self.controlArea, "Variables")
        opts = dict(
            widget=self.options_box, master=self, orientation=Qt.Horizontal,
            callback=self.change_attributes)
        gui.comboBox(value='attr_x', model=self.var_model, **opts)
        gui.comboBox(value='attr_y', model=self.var_model, **opts)
        gui.checkBox(self.options_box, self, 'normalization',
                     'Scale and center data',
                     callback=self.change_attributes)
        gui.comboBox(value='target_class', label='Target: ',
                     sendSelectedValue=True, **opts)

        self.properties_box = gui.widgetBox(self.controlArea, "Properties")
        self.alpha_spin = gui.spin(
            widget=self.properties_box, master=self, callback=self.change_alpha,
            value="alpha", label="Learning rate: ",
            minv=0.001, maxv=100, step=0.001, spinType=float, decimals=3,
            alignment=Qt.AlignRight, controlWidth=80)
        self.stochastic_checkbox = gui.checkBox(
            widget=self.properties_box, master=self,
            callback=self.change_stochastic, value="stochastic",
            label="Stochastic descent")
        self.step_size_spin = gui.spin(
            widget=self.properties_box, master=self, callback=self.change_step,
            value="step_size", label="Step size: ",
            minv=1, maxv=100, step=1, alignment=Qt.AlignRight, controlWidth=80)

        self.step_box = gui.widgetBox(self.controlArea, box=True,
                                      margin=0, spacing=0)
        hbox = gui.hBox(self.step_box)
        self.step_button = gui.button(
            hbox, self, callback=self.step, label="Step", default=True)
        self.step_back_button = gui.button(
            hbox, self, callback=self.step_back, label="Back")
        self.restart_button = gui.button(
            widget=self.step_box, master=self,
            callback=self.restart, label="Restart")

        gui.separator(self.step_box)
        self.auto_play_button = gui.button(
            widget=self.step_box, master=self,
            label=self.auto_play_button_text[0], callback=self.auto_play)

        gui.rubber(self.controlArea)

        self._add_graph()

        self.step_size_lock()
        self.step_back_button_lock()

    def _add_graph(self):
        self.graph = pg.PlotWidget(background="w", autoRange=False)
        self.graph.setCursor(QCursor(Qt.CrossCursor))
        self._hover_delegate = HoverEventDelegate(self.help_event)
        self.graph.scene().installEventFilter(self._hover_delegate)
        self.mainArea.layout().addWidget(self.graph)

        self.hover_label = label = QGraphicsTextItem()
        label.setFlag(QGraphicsTextItem.ItemIgnoresTransformations)
        self.graph.addItem(label)
        label.setZValue(10)
        label.hide()
        # I'm not proud of this and will brew a coffee to the person who
        # improves it (and comes to claim it). Same in Polynomial Classification
        self.graph.scene().leaveEvent = lambda *_: self.hover_label.hide()

        self.optimization_label = label = QGraphicsTextItem()
        label.setFlag(QGraphicsTextItem.ItemIgnoresTransformations)
        self.graph.addItem(label)
        label.setZValue(10)

    ##############################
    # User-interaction

    def change_attributes(self):
        self.attr_x_hint = self.attr_x.name if self.attr_x else ""
        self.attr_y_hint = self.attr_y.name if self.attr_y else ""
        self.target_class_hint = self.target_class
        self.set_axis_titles()
        self.learner = None  # that theta does not same equal
        self.restart()

    def change_alpha(self):
        if self.learner is not None:
            self.learner.set_alpha(self.alpha)

    def change_stochastic(self):
        if self.learner is not None:
            self.learner.stochastic = self.stochastic
        self.step_size_lock()

    def change_step(self):
        if self.learner is not None:
            self.learner.stochastic_step_size = self.step_size

    def change_theta(self, x, y):
        if self.learner is None:
            return
        self.learner.set_theta([x, y])
        self.update_history()

    def step(self):
        if self.data is None:
            return
        if self.learner.step_no > 500:  # limit step no to avoid freezes
            return
        self.learner.step()
        self.update_history()
        self.send_output()
        self.step_back_button_lock()

    def step_back(self):
        if self.data is None:
            return
        if self.learner.step_no > 0:
            self.learner.step_back()
            self.update_history()
            self.send_output()
        self.step_back_button_lock()

    def restart(self):
        self.clear_plot()
        self.select_columns()
        self.set_axis_titles()
        if self.selected_data is None:
            return

        theta = self.learner.history[0][0] if self.learner is not None else None
        selected_learner = (LogisticRegression if self.is_classification
                            else LinearRegression)
        self.learner = selected_learner(
            data=self.selected_data,
            alpha=self.alpha, stochastic=self.stochastic,
            theta=theta, step_size=self.step_size,
            intercept=not self.is_classification)
        self.replot()
        if theta is None:  # no previous theta exist
            self.change_theta(np.random.uniform(self.min_x, self.max_x),
                              np.random.uniform(self.min_y, self.max_y))
        else:
            self.change_theta(theta[0], theta[1])
        self.send_output()
        self.step_back_button_lock()

    def auto_play(self):
        if self.data is not None:
            self.auto_play_enabled = not self.auto_play_enabled
            self.auto_play_button.setText(
                self.auto_play_button_text[self.auto_play_enabled])
            if self.auto_play_enabled:
                self.disable_controls(self.auto_play_enabled)
                self.auto_play_thread = Autoplay(self)
                self.step_trigger.connect(self.step)
                self.stop_auto_play_trigger.connect(self.stop_auto_play)
                self.auto_play_thread.start()
            else:
                self.stop_auto_play()

    def stop_auto_play(self):
        self.auto_play_enabled = False
        self.disable_controls(self.auto_play_enabled)
        self.auto_play_button.setText(
            self.auto_play_button_text[self.auto_play_enabled])

    def step_back_button_lock(self):
        self.step_back_button.setDisabled(
            self.learner is None
            or self.learner.step_no == 0
            or self.auto_play_enabled)

    def step_size_lock(self):
        self.step_size_spin.setDisabled(not self.stochastic)

    def disable_controls(self, disabled):
        for item in [self.step_button, self.step_back_button,
                     self.restart_button, self.properties_box, self.options_box]:
            item.setDisabled(disabled)

    key_actions = {(Qt.NoModifier, Qt.Key_Space): step}  # space button for step

    def keyPressEvent(self, e):
        """Bind 'back' key to step back"""
        if (e.modifiers(), e.key()) in self.key_actions:
            fun = self.key_actions[(e.modifiers(), e.key())]
            fun(self)
        else:
            super().keyPressEvent(e)

    ##############################
    # Signals and data-handling

    @Inputs.data
    def set_data(self, data):
        target_combo = self.controls.target_class
        self.Error.clear()

        self.cost_grid = None
        self.learner = None
        self.selected_data = None
        self.data = None
        domain = None  # for pylint
        target_combo.clear()
        self.clear_plot()

        if data:
            domain = data.domain
            self.var_model[:] = [
                attr for attr in domain.attributes if attr.is_continuous]
            if data.domain.class_var is None:
                data = None
                self.Error.no_class()
            # don't use self.is_classification -- widget has no self.data yet
            elif domain.class_var.is_discrete:
                if len(domain.class_var.values) < 2:
                    self.Error.no_class_values()
                    data = None
                elif len(self.var_model) < 2:
                    self.Error.num_features("two numeric variables")
                    data = None
            elif len(self.var_model) < 1:
                self.Error.num_features("one numeric variable")
                data = None

        if not data:
            self.var_model.clear()
            self.send_output()
            return

        assert domain is not None
        self.data = data
        self.controls.attr_y.setHidden(not self.is_classification)
        target_combo.box.setHidden(not self.is_classification)
        attr_names = [var.name for var in self.var_model]
        if self.attr_x_hint in attr_names:
            self.attr_x = domain[self.attr_x_hint]
        else:
            self.attr_x = self.var_model[0]
            self.attr_x_hint = self.attr_x.name
        if self.is_classification:
            if self.attr_y_hint in attr_names:
                self.attr_y = domain[self.attr_y_hint]
            else:
                self.attr_y = self.var_model[int(self.attr_x == self.var_model[0])]
                self.attr_y_hint = self.attr_y.name
            target_combo.clear()
            values = data.domain.class_var.values
            target_combo.addItems(values)
            # For binary class, take value with index 1 as a target, so the
            # output class is the same as the inputđ
            if self.target_class_hint in values:
                self.target_class = self.target_class_hint
            else:
                self.target_class = values[1 if len(values) == 2 else 0]
                self.target_class_hint = self.target_class
        self.step_size_spin.setMaximum(len(data))
        self.set_axis_titles()
        self.restart()

    @property
    def is_classification(self):
        return self.data and self.data.domain.class_var.is_discrete

    def select_columns(self):
        self.Error.no_nonnan_data.clear()
        self.Error.same_variable.clear()
        self.selected_data = None
        if self.data is None:
            return

        old_class = self.data.domain.class_var
        if self.is_classification:
            if self.attr_x is self.attr_y:
                self.Error.same_variable()
                return
            values = old_class.values
            target_idx = values.index(self.target_class)
            new_class = DiscreteVariable(
                old_class.name + "'",
                values=(values[1 - target_idx] if len(values) == 2 else 'Others',
                        self.target_class),
                compute_value=Indicator(old_class, target_idx))
            domain = Domain([self.attr_x, self.attr_y], new_class, [old_class])
        else:
            domain = Domain([self.attr_x], old_class)

        data = self.data.transform(domain)
        valid_data = \
            np.flatnonzero(
                np.all(
                    np.isfinite(data.X),
                    axis=1)
            )

        if not valid_data.size:
            self.Error.no_nonnan_data()
            return

        data = data[valid_data]
        if self.normalization:
            self.selected_data = Normalize(
                transform_class=not self.is_classification)(
                data)
        else:
            self.selected_data = data

    def send_output(self):
        self.send_model()
        self.send_coefficients()

    def send_model(self):
        if self.learner is not None and self.learner.theta is not None:
            self.Outputs.model.send(self.learner.model)
        else:
            self.Outputs.model.send(None)

    def send_coefficients(self):
        if self.learner is None or self.learner.theta is None:
            self.Outputs.coefficients.send(None)
            return

        domain = Domain(
            [ContinuousVariable("Coefficient")],
            metas=[StringVariable("Variable")])
        if self.is_classification:
            names = [self.attr_x.name, self.attr_y.name]
        else:
            names = ["intercept", self.attr_x.name]

        coefficients_table = Table.from_list(
            domain, list(zip(list(self.learner.theta), names)))
        self.Outputs.coefficients.send(coefficients_table)

    def send_report(self):
        if self.data is None:
            return
        caption_items = (
            ("Target class", self.target_class),
            ("Learning rate", self.alpha),
            ("Stochastic", str(self.stochastic))
        )
        if self.stochastic:
            caption_items += (("Stochastic step size", self.step_size),)
        caption = report.render_items_vert(caption_items)
        self.report_plot()
        self.report_caption(caption)

    ##############################
    # Plot-related methods

    def clear_plot(self):
        self.graph.clear()
        self.graph.addItem(self.hover_label)
        self.graph.addItem(self.optimization_label)
        self.optimization_label.hide()

    def replot(self):
        if self.data is None or self.selected_data is None:
            self.clear_plot()
            return

        optimal_theta = self.learner.optimized()
        self.min_x = optimal_theta[0] - 10
        self.max_x = optimal_theta[0] + 10
        self.min_y = optimal_theta[1] - 10
        self.max_y = optimal_theta[1] + 10
        self.graph.setRange(xRange=(self.min_x, self.max_x),
                            yRange=(self.min_y, self.max_y))

        x = np.linspace(self.min_x, self.max_x, GRID_SIZE)
        y = np.linspace(self.min_y, self.max_y, GRID_SIZE)
        xv, yv = np.meshgrid(x, y)
        thetas = np.column_stack((xv.flatten(), yv.flatten()))
        cost_values = self.learner.j(thetas)
        self.cost_grid = cost_values.reshape(xv.shape)

        self.plot_gradient()
        self.plot_contour(xv, yv)
        self.add_paths()

    def plot_gradient(self):
        if self.cost_grid is None:
            return

        cg = interpolate_grid(self.cost_grid, 256)
        bitmap = cg * (255 / np.max(cg))  # make a copy
        bitmap = bitmap.astype(np.uint8)

        h, s, v = rgb_to_hsv(*np.array(self.default_background_color) / 255)
        palette = np.linspace([h, 0, v], [h, s, 1], 255)
        palette = 255 * np.array([hsv_to_rgb(*col) for col in palette])
        palette = palette.astype(int)

        density_img = pg.ImageItem(bitmap.T, lut=palette)
        density_img.setRect(
            QRectF(self.min_x, self.min_y,
                   self.max_x - self.min_x, self.max_y - self.min_y))
        density_img.setZValue(-1)
        self.graph.addItem(density_img, ignoreBounds=True)

    def plot_contour(self, xv, yv):
        while self.contours:
            self.graph.removeItem(self.contours.pop())

        if self.cost_grid is None:
            return

        contour = Contour(xv, yv, self.cost_grid)
        contour_lines = contour.contours(
            np.linspace(np.min(self.cost_grid), np.max(self.cost_grid), 20))
        for key, value in contour_lines.items():
            for line in value:
                if len(line) > 3:
                    tck, u = splprep(np.array(line).T, s=0.0, per=0)
                    u_new = np.linspace(u.min(), u.max(), 100)
                    x_new, y_new = splev(u_new, tck, der=0)
                    interpol_line = np.c_[x_new, y_new]
                else:
                    interpol_line = line

                contour = pg.PlotCurveItem(
                    *np.array(list(interpol_line)).T,
                    pen=self._contour_pen(False),
                    antialias=True
                )
                contour.value = key
                self.graph.addItem(contour)
                self.contours.append(contour)

    def add_paths(self):
        yellow = QColor(255, 165, 0)
        dark = yellow.darker()
        self.optimization_path = pg.PlotDataItem(
            [], [], pen=pg.mkPen(dark, width=2),
            symbolPen=pg.mkPen(yellow), symbolBrush=pg.mkBrush(dark),
            symbol="o", symbolSize=4)
        self.last_point = pg.PlotDataItem(
            [], [],
            symbolPen=pg.mkPen(yellow), symbolBrush=pg.mkBrush(yellow),
            symbol="o", symbolSize=7)
        self.graph.addItem(self.optimization_path)
        self.graph.addItem(self.last_point)

    @staticmethod
    def _contour_pen(is_hovered=False):
        return pg.mkPen(0.2, width=1 + 2 * is_hovered)

    def set_axis_titles(self):
        if self.selected_data is None:
            names = ["", ""]
        elif self.is_classification:
            names = [f"Θ({self.attr_x.name})", f"Θ({self.attr_y.name})"]
        else:
            names = ["Θ₀", "Θ₁"]
        for axname, name in zip(("bottom", "left"), names):
            axis = self.graph.getAxis(axname)
            axis.setLabel(name)

    def update_history(self):
        steps = self.learner.step_no
        x, y = zip(*(coords for coords, *_ in self.learner.history[:steps + 1]))
        self.optimization_path.setData(x, y, data=np.arange(steps + 1))

        lx, ly = x[-1], y[-1]
        self.last_point.setData([lx], [ly], shape="o")
        self.optimization_label.setPos(lx, ly)
        self.optimization_label.setHtml(self._format_label(lx, ly, steps))
        self.optimization_label.show()

        self._update_hover_visibility()

        self.send_output()

    ##############################
    # Plot labels

    def help_event(self, event):
        if self.cost_grid is None:
            return False

        pos = event.scenePos()
        pos = self.graph.mapToView(pos)
        xc, yc = pos.x(), pos.y()

        if event.button() == Qt.LeftButton:
            self.change_theta(xc, yc)
            return True

        label = self.hover_label
        hovered_line = None

        # In case there are multiple hovered points or lines, pick the middle
        opt_points = self.optimization_path.scatter.pointsAt(pos)
        if opt_points.size:
            point = opt_points[len(opt_points) // 2]
            ppos = point.pos()
            label.setHtml(self._format_label(ppos.x(), ppos.y(), point.data()))
        else:
            hovereds = [item for item in self.contours
                        if isinstance(item, pg.PlotCurveItem)
                        and item.mouseShape().contains(pos)]
            if hovereds:
                hovered_line = hovereds[len(hovereds) // 2]
                # Show the cost for the line at mouse position, not the one
                # computed from coordinates. Use a different format to
                # distinguish this "line label" from the usual hover label
                label.setHtml(f"<b>Cost: {hovered_line.value:.3f}</b>")
            else:
                cost = self.learner.j(np.array([xc, yc]))
                label.setHtml(f"{xc:.3f}, {yc:.3f}<br/>Cost: {cost:.5f}")

        # Set the pen for all lines, hovered and not hovered (any longer)
        for item in self.contours:
            item.setPen(self._contour_pen(item is hovered_line))

        label.setPos(pos)
        label.show()
        self._update_hover_visibility()
        return True

    def _update_hover_visibility(self):
        # QGraphicsTextItem.collidesWith doesn't seem to work. I'm stupid,
        # or it may be related to ignoring transformations.
        opt_label = self.optimization_label
        hover_label = self.hover_label
        bopt = opt_label.boundingRect()
        bopt.moveTo(self.graph.mapFromView(opt_label.pos()))
        bhov = hover_label.boundingRect()
        bhov.moveTo(self.graph.mapFromView(hover_label.pos()))
        if hover_label.isVisible() and bhov.intersects(bopt):
            opt_label.hide()
        else:
            opt_label.show()

    def _format_label(self, x, y, step):
        return \
            f"<b>Step {step}:</b><br/>" \
            f"{x:.3f}, {y:.3f}<br/>" \
            f"Cost: {self.learner.j(np.array([x, y])):.5f}"

    def changeEvent(self, ev):
        # This hides the label if the user alt-tabs out of the window
        if ev.type() == QEvent.ActivationChange and not self.isActiveWindow():
            self.hover_label.hide()
            self._update_hover_visibility()
        super().changeEvent(ev)


if __name__ == "__main__":
    WidgetPreview(OWGradientDescent).run(Table.from_file('iris'))
    # WidgetPreview(OWGradientDescent).run(Table.from_file('housing'))
